###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Initial property package for multi-ionic system for use in the
Donnan Steric Pore Model with Dielectric Exclusion (DSPM-DE)
"""

# TODO:
#  -add calc option for Stokes radius from Stokes Einstein
#  -add viscosity as func of temp and concentration

# Import Python libraries
import idaes.logger as idaeslog

from enum import Enum, auto
# Import Pyomo libraries
from pyomo.environ import Constraint, Expression, Reals, NonNegativeReals, log, \
    Var, Param, Suffix, value, check_optimal_termination, units as pyunits
from pyomo.common.config import ConfigValue, In

# Import IDAES cores
from idaes.core import (declare_process_block_class,
                        MaterialFlowBasis,
                        PhysicalParameterBlock,
                        StateBlockData,
                        StateBlock,
                        MaterialBalanceType,
                        )
from idaes.core.components import Component, Solute, Solvent
from idaes.core.phases import LiquidPhase
from idaes.core.util.constants import Constants
from idaes.core.util.initialization import (fix_state_vars,
                                            revert_state_vars,
                                            solve_indexed_blocks)
from idaes.core.util.misc import add_object_reference, extract_data
from idaes.core.util import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom, \
    number_unfixed_variables
from idaes.core.util.exceptions import ConfigurationError, InitializationError
import idaes.core.util.scaling as iscale

# Set up logger
_log = idaeslog.getLogger(__name__)


class ActivityCoefficientModel(Enum):
    ideal = auto()                    # Ideal
    davies = auto()                   # Davies

class DensityCalculation(Enum):
    constant = auto()                 # constant @ 1000 kg/m3
    seawater = auto()                 # seawater correlation for TDS from Sharqawy
    laliberte = auto()                # Laliberte correlation using apparent density #TODO add this later with reference


@declare_process_block_class("DSPMDEParameterBlock")
class DSPMDEParameterData(PhysicalParameterBlock):
    CONFIG = PhysicalParameterBlock.CONFIG()

    CONFIG.declare("solute_list", ConfigValue(
        domain=list,
        description="List of solute species names"))
    CONFIG.declare("stokes_radius_data", ConfigValue(
        default={},
        domain=dict,
        description="Dict of solute species names and Stokes radius data"))
    CONFIG.declare("diffusivity_data", ConfigValue(
        default={},
        domain=dict,
        description="Dict of solute species names and bulk ion diffusivity data"))
    CONFIG.declare("mw_data", ConfigValue(
        default={},
        domain=dict,
        description="Dict of component names and molecular weight data"))
    CONFIG.declare("charge", ConfigValue(
        default={},
        domain=dict,
        description="Ion charge"))
    CONFIG.declare("activity_coefficient_model", ConfigValue(
        default=ActivityCoefficientModel.ideal,
        domain=In(ActivityCoefficientModel),
        description="Activity coefficient model construction flag",
        doc="""
           Options to account for activity coefficient model.
    
           **default** - ``ActivityCoefficientModel.ideal``
    
       .. csv-table::
           :header: "Configuration Options", "Description"
    
           "``ActivityCoefficientModel.ideal``", "Activity coefficients equal to 1 assuming ideal solution"
           "``ActivityCoefficientModel.davies``", "Activity coefficients estimated via Davies model"
       """))
    CONFIG.declare("density_calculation", ConfigValue(
        default=DensityCalculation.constant,
        domain=In(DensityCalculation),
        description="Solution density calculation construction flag",
        doc="""
           Options to account for solution density.
    
           **default** - ``DensityCalculation.seawater``
    
       .. csv-table::
           :header: "Configuration Options", "Description"
    
           "``DensityCalculation.constant``", "Solution density assumed constant at 1000 kg/m3"
           "``DensityCalculation.seawater``", "Solution density based on correlation for seawater (TDS)"
           "``DensityCalculation.laliberte``", "Solution density based on mixing correlation from Laliberte"
       """))


    def build(self):
        '''
        Callable method for Block construction.
        '''
        super().build()


        self._state_block_class = DSPMDEStateBlock

        # components
        self.H2O = Solvent()

        for j in self.config.solute_list:
            self.add_component(str(j), Solute())

        # phases
        self.Liq = LiquidPhase()

        # reference
        # Todo: enter any relevant references

        # TODO: consider turning parameters into variables for future param estimation
        # molecular weight
        self.mw_comp = Param(
            self.component_list,
            mutable=True,
            default=18e-3,
            initialize=self.config.mw_data,
            units=pyunits.kg/pyunits.mol,
            doc="Molecular weight")
        # Stokes radius
        self.radius_stokes_comp = Param(
            self.solute_set,
            mutable=True,
            default=1e-10,
            initialize=self.config.stokes_radius_data,
            units=pyunits.m,
            doc="Stokes radius of solute")
        self.diffus_phase_comp = Param(
            self.phase_list,
            self.solute_set,
            mutable=True,
            default=1e-9,
            initialize=self.config.diffusivity_data,
            units=pyunits.m ** 2 * pyunits.s ** -1,
            doc="Bulk diffusivity of ion")
        self.visc_d_phase = Param(
            self.phase_list,
            mutable=True,
            default=1e-3,
            initialize=1e-3, #TODO:revisit- assuming ~ 1e-3 Pa*s for pure water
            units=pyunits.Pa * pyunits.s,
            doc="Fluid viscosity")
        # Ion charge
        self.charge_comp = Param(
            self.solute_set,
            mutable=True,
            default=1,
            initialize=self.config.charge,
            units=pyunits.dimensionless,
            doc="Ion charge")
        # Dielectric constant of water
        self.dielectric_constant = Param(
            mutable=True,
            default=80.4,
            initialize=80.4, #todo: make a variable with parameter values for coefficients in the function of temperature
            units=pyunits.dimensionless,
            doc="Dielectric constant of water")
        self.debye_huckel_b = Param(
            mutable=True,
            default=0.3,
            initialize=0.3,
            units=pyunits.kg/pyunits.mol,
            doc="Debye Huckel constant b")

        # Mass density parameters, eq. 8 in Sharqawy et al. (2010)
        dens_units = pyunits.kg / pyunits.m ** 3
        t_inv_units = pyunits.K ** -1

        self.dens_mass_param_A1 = Var(
            within=Reals, initialize=9.999e2, units=dens_units,
            doc='Mass density parameter A1')
        self.dens_mass_param_A2 = Var(
            within=Reals, initialize=2.034e-2, units=dens_units * t_inv_units,
            doc='Mass density parameter A2')
        self.dens_mass_param_A3 = Var(
            within=Reals, initialize=-6.162e-3, units=dens_units * t_inv_units ** 2,
            doc='Mass density parameter A3')
        self.dens_mass_param_A4 = Var(
            within=Reals, initialize=2.261e-5, units=dens_units * t_inv_units ** 3,
            doc='Mass density parameter A4')
        self.dens_mass_param_A5 = Var(
            within=Reals, initialize=-4.657e-8, units=dens_units * t_inv_units ** 4,
            doc='Mass density parameter A5')
        self.dens_mass_param_B1 = Var(
            within=Reals, initialize=8.020e2, units=dens_units,
            doc='Mass density parameter B1')
        self.dens_mass_param_B2 = Var(
            within=Reals, initialize=-2.001, units=dens_units * t_inv_units,
            doc='Mass density parameter B2')
        self.dens_mass_param_B3 = Var(
            within=Reals, initialize=1.677e-2, units=dens_units * t_inv_units ** 2,
            doc='Mass density parameter B3')
        self.dens_mass_param_B4 = Var(
            within=Reals, initialize=-3.060e-5, units=dens_units * t_inv_units ** 3,
            doc='Mass density parameter B4')
        self.dens_mass_param_B5 = Var(
            within=Reals, initialize=-1.613e-5, units=dens_units * t_inv_units ** 2,
            doc='Mass density parameter B5')

        # traditional parameters are the only Vars currently on the block and should be fixed
        for v in self.component_objects(Var):
            v.fix()

        # ---default scaling---
        self.set_default_scaling('temperature', 1e-2)
        self.set_default_scaling('pressure', 1e-6)
        self.set_default_scaling('dens_mass_phase', 1e-3, index='Liq')
        self.set_default_scaling('visc_d_phase', 1e3, index='Liq')
        self.set_default_scaling('diffus_phase_comp', 1e10, index='Liq')



    @classmethod
    def define_metadata(cls, obj):
        """Define properties supported and units."""
        obj.add_properties(
            {'flow_mol_phase_comp': {'method': None},
             'temperature': {'method': None},
             'pressure': {'method': None},
             'flow_mass_phase_comp': {'method': '_flow_mass_phase_comp'},
             'mass_frac_phase_comp': {'method': '_mass_frac_phase_comp'},
             'dens_mass_phase': {'method': '_dens_mass_phase'},
             'dens_mass_solvent': {'method': '_dens_mass_solvent'},
             'flow_vol': {'method': '_flow_vol'},
             'flow_vol_phase': {'method': '_flow_vol_phase'},
             'conc_mol_phase_comp': {'method': '_conc_mol_phase_comp'},
             'conc_mass_phase_comp': {'method': '_conc_mass_phase_comp'},
             'mole_frac_phase_comp': {'method': '_mole_frac_phase_comp'},
             'molality_comp': {'method': '_molality_comp'},
             'diffus_phase_comp': {'method': '_diffus_phase_comp'},
             'visc_d_phase': {'method': '_visc_d_phase'},
             'pressure_osm': {'method': '_pressure_osm'},
             'radius_stokes_comp': {'method': '_radius_stokes_comp'},
             'mw_comp': {'method': '_mw_comp'},
             'charge_comp': {'method': '_charge_comp'},
             'act_coeff_phase_comp': {'method': '_act_coeff_phase_comp'},
             'dielectric_constant': {'method': '_dielectric_constant'},
             'debye_huckel_constant': {'method': '_debye_huckel_constant'},
             'ionic_strength': {'method': '_ionic_strength'}
             })

        obj.add_default_units({'time': pyunits.s,
                               'length': pyunits.m,
                               'mass': pyunits.kg,
                               'amount': pyunits.mol,
                               'temperature': pyunits.K})


class _DSPMDEStateBlock(StateBlock):
    """
    This Class contains methods which should be applied to Property Blocks as a
    whole, rather than individual elements of indexed Property Blocks.
    """

    def initialize(self, state_args=None, state_vars_fixed=False,
                   hold_state=False, outlvl=idaeslog.NOTSET,
                   solver=None, optarg=None):
        """
        Initialization routine for property package.
        Keyword Arguments:
            state_args : Dictionary with initial guesses for the state vars
                         chosen. Note that if this method is triggered
                         through the control volume, and if initial guesses
                         were not provided at the unit model level, the
                         control volume passes the inlet values as initial
                         guess.The keys for the state_args dictionary are:

                         flow_mol_phase_comp : value at which to initialize
                                               phase component flows
                         pressure : value at which to initialize pressure
                         temperature : value at which to initialize temperature
            outlvl : sets output level of initialization routine (default=idaeslog.NOTSET)
            optarg : solver options dictionary object (default=None)
            state_vars_fixed: Flag to denote if state vars have already been
                              fixed.
                              - True - states have already been fixed by the
                                       control volume 1D. Control volume 0D
                                       does not fix the state vars, so will
                                       be False if this state block is used
                                       with 0D blocks.
                             - False - states have not been fixed. The state
                                       block will deal with fixing/unfixing.
            solver : Solver object to use during initialization if None is provided
                     it will use the default solver for IDAES (default = None)
            hold_state : flag indicating whether the initialization routine
                         should unfix any state variables fixed during
                         initialization (default=False).
                         - True - states variables are not unfixed, and
                                 a dict of returned containing flags for
                                 which states were fixed during
                                 initialization.
                        - False - state variables are unfixed after
                                 initialization by calling the
                                 release_state method
        Returns:
            If hold_states is True, returns a dict containing flags for
            which states were fixed during initialization.
        """
        # Get loggers
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="properties")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="properties")

        # Set solver and options
        opt = get_solver(solver, optarg)

        # Fix state variables
        flags = fix_state_vars(self, state_args)
        # Check when the state vars are fixed already result in dof 0
        for k in self.keys():
            dof = degrees_of_freedom(self[k])
            if dof != 0:
                raise InitializationError("\nWhile initializing {sb_name}, the degrees of freedom "
                                           "are {dof}, when zero is required. \nInitialization assumes "
                                           "that the state variables should be fixed and that no other "
                                           "variables are fixed. \nIf other properties have a "
                                           "predetermined value, use the calculate_state method "
                                           "before using initialize to determine the values for "
                                           "the state variables and avoid fixing the property variables."
                                           "".format(sb_name=self.name, dof=dof))

        # ---------------------------------------------------------------------
        skip_solve = True  # skip solve if only state variables are present
        for k in self.keys():
            if number_unfixed_variables(self[k]) != 0:

                skip_solve = False

        if not skip_solve:
            # Initialize properties
            with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
                results = solve_indexed_blocks(opt, [self], tee=slc.tee)
                if not check_optimal_termination(results):
                    raise InitializationError('The property package failed to solve during initialization.')
            init_log.info_high("Property initialization: {}.".format(idaeslog.condition(results)))

        # ---------------------------------------------------------------------
        # If input block, return flags, else release state
        if state_vars_fixed is False:
            if hold_state is True:
                return flags
            else:
                self.release_state(flags)


    def release_state(self, flags, outlvl=idaeslog.NOTSET):
        '''
        Method to release state variables fixed during initialisation.

        Keyword Arguments:
            flags : dict containing information of which state variables
                    were fixed during initialization, and should now be
                    unfixed. This dict is returned by initialize if
                    hold_state=True.
            outlvl : sets output level of logging
        '''
        # Unfix state variables
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="properties")
        revert_state_vars(self, flags)
        init_log.info_high('{} State Released.'.format(self.name))

    def calculate_state(self, var_args=None, hold_state=False, outlvl=idaeslog.NOTSET,
                        solver=None, optarg=None):
        """
        Solves state blocks given a set of variables and their values. These variables can
        be state variables or properties. This method is typically used before
        initialization to solve for state variables because non-state variables (i.e. properties)
        cannot be fixed in initialization routines.

        Keyword Arguments:
            var_args : dictionary with variables and their values, they can be state variables or properties
                       {(VAR_NAME, INDEX): VALUE}
            hold_state : flag indicating whether all of the state variables should be fixed after calculate state.
                         True - State variables will be fixed.
                         False - State variables will remain unfixed, unless already fixed.
            outlvl : idaes logger object that sets output level of solve call (default=idaeslog.NOTSET)
            solver : solver name string if None is provided the default solver
                     for IDAES will be used (default = None)
            optarg : solver options dictionary object (default={})

        Returns:
            results object from state block solve
        """
        # Get logger
        solve_log = idaeslog.getSolveLogger(self.name, level=outlvl, tag="properties")

        # Initialize at current state values (not user provided)
        self.initialize(solver=solver, optarg=optarg, outlvl=outlvl)

        # Set solver and options
        opt = get_solver(solver, optarg)

        # Fix variables and check degrees of freedom
        flags = {}  # dictionary noting which variables were fixed and their previous state
        for k in self.keys():
            sb = self[k]
            for (v_name, ind), val in var_args.items():
                var = getattr(sb, v_name)
                if iscale.get_scaling_factor(var[ind]) is None:
                    _log.warning(
                            "While using the calculate_state method on {sb_name}, variable {v_name} "
                            "was provided as an argument in var_args, but it does not have a scaling "
                            "factor. This suggests that the calculate_scaling_factor method has not been "
                            "used or the variable was created on demand after the scaling factors were "
                            "calculated. It is recommended to touch all relevant variables (i.e. call "
                            "them or set an initial value) before using the calculate_scaling_factor "
                            "method.".format(v_name=v_name, sb_name=sb.name))
                if var[ind].is_fixed():
                    flags[(k, v_name, ind)] = True
                    if value(var[ind]) != val:
                        raise ConfigurationError(
                            "While using the calculate_state method on {sb_name}, {v_name} was "
                            "fixed to a value {val}, but it was already fixed to value {val_2}. "
                            "Unfix the variable before calling the calculate_state "
                            "method or update var_args."
                            "".format(sb_name=sb.name, v_name=var.name, val=val, val_2=value(var[ind])))
                else:
                    flags[(k, v_name, ind)] = False
                    var[ind].fix(val)

            if degrees_of_freedom(sb) != 0:
                raise RuntimeError("While using the calculate_state method on {sb_name}, the degrees "
                                   "of freedom were {dof}, but 0 is required. Check var_args and ensure "
                                   "the correct fixed variables are provided."
                                   "".format(sb_name=sb.name, dof=degrees_of_freedom(sb)))

        # Solve
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            results = solve_indexed_blocks(opt, [self], tee=slc.tee)
            solve_log.info_high("Calculate state: {}.".format(idaeslog.condition(results)))

        if not check_optimal_termination(results):
            _log.warning("While using the calculate_state method on {sb_name}, the solver failed "
                         "to converge to an optimal solution. This suggests that the user provided "
                         "infeasible inputs, or that the model is poorly scaled, poorly initialized, "
                         "or degenerate.")

        # unfix all variables fixed with var_args
        for (k, v_name, ind), previously_fixed in flags.items():
            if not previously_fixed:
                var = getattr(self[k], v_name)
                var[ind].unfix()

        # fix state variables if hold_state
        if hold_state:
            fix_state_vars(self)

        return results

@declare_process_block_class("DSPMDEStateBlock",
                             block_class=_DSPMDEStateBlock)
class DSPMDEStateBlockData(StateBlockData):
    def build(self):
        """Callable method for Block construction."""
        super().build()

        self.scaling_factor = Suffix(direction=Suffix.EXPORT)

        # Add state variables
        self.flow_mol_phase_comp = Var(
            self.params.phase_list,
            self.params.component_list,
            initialize=100, #todo: revisit
            bounds=(1e-8, None),
            domain=NonNegativeReals,
            units=pyunits.mol/pyunits.s,
            doc='Mole flow rate')

        self.temperature = Var(
            initialize=298.15,
            bounds=(273.15, 373.15),
            domain=NonNegativeReals,
            units=pyunits.K,
            doc='State temperature')

        self.pressure = Var(
            initialize=101325,
            bounds=(1e5, 5e7),
            domain=NonNegativeReals,
            units=pyunits.Pa,
            doc='State pressure')

    # -----------------------------------------------------------------------------
    # Property Methods
    def _mass_frac_phase_comp(self):
        self.mass_frac_phase_comp = Var(
            self.params.phase_list,
            self.params.component_list,
            initialize=lambda b,p,j : 0.4037 if j == "H2O" else 0.0033, #todo: revisit
            bounds=(1e-8, 1.001),
            units=pyunits.kg/pyunits.kg,
            doc='Mass fraction')

        def rule_mass_frac_phase_comp(b, j):
            return (b.mass_frac_phase_comp['Liq', j] == b.flow_mass_phase_comp['Liq', j] /
                    sum(b.flow_mass_phase_comp['Liq', j]
                        for j in self.params.component_list))
        self.eq_mass_frac_phase_comp = Constraint(self.params.component_list, rule=rule_mass_frac_phase_comp)

    def _dens_mass_phase(self):
        self.dens_mass_phase = Var(
            ['Liq'],
            initialize=1e3,
            bounds=(5e2, 2e3),
            units=pyunits.kg * pyunits.m ** -3,
            doc="Mass density")
        #TODO: reconsider this approach for solution density based on arbitrary solute_list
        def rule_dens_mass_phase(b):
            if b.params.config.density_calculation == DensityCalculation.constant:
                return (b.dens_mass_phase['Liq'] == 1000 * pyunits.kg * pyunits.m**-3)
            elif b.params.config.density_calculation == DensityCalculation.seawater:
                # density, eq. 8 in Sharqawy #TODO- add Sharqawy reference
                t = b.temperature - 273.15 * pyunits.K
                s = sum(b.mass_frac_phase_comp['Liq', j] for j in b.params.solute_set)
                dens_mass = (b.dens_mass_solvent
                             + b.params.dens_mass_param_B1 * s
                             + b.params.dens_mass_param_B2 * s * t
                             + b.params.dens_mass_param_B3 * s * t ** 2
                             + b.params.dens_mass_param_B4 * s * t ** 3
                             + b.params.dens_mass_param_B5 * s ** 2 * t ** 2)
                return b.dens_mass_phase['Liq'] == dens_mass
        self.eq_dens_mass_phase = Constraint(rule=rule_dens_mass_phase)

    def _dens_mass_solvent(self):
        self.dens_mass_solvent = Var(
            initialize=1e3,
            bounds=(1, 1e6),
            units=pyunits.kg*pyunits.m**-3,
            doc="Mass density of pure water")

        def rule_dens_mass_solvent(b):  # density, eq. 8 in Sharqawy
            t = b.temperature - 273.15*pyunits.K
            dens_mass_w = (b.params.dens_mass_param_A1
                         + b.params.dens_mass_param_A2 * t
                         + b.params.dens_mass_param_A3 * t**2
                         + b.params.dens_mass_param_A4 * t**3
                         + b.params.dens_mass_param_A5 * t**4)
            return b.dens_mass_solvent == dens_mass_w
        self.eq_dens_mass_solvent = Constraint(rule=rule_dens_mass_solvent)

    def _flow_vol_phase(self):
        self.flow_vol_phase = Var(
            self.params.phase_list,
            initialize=1,
            bounds=(1e-8, None),
            units=pyunits.m ** 3 / pyunits.s,
            doc="Volumetric flow rate")

        def rule_flow_vol_phase(b):
            return (b.flow_vol_phase['Liq']
                    == sum(b.flow_mol_phase_comp['Liq', j]*b.mw_comp[j] for j in self.params.component_list)
                    / b.dens_mass_phase['Liq'])
        self.eq_flow_vol_phase = Constraint(rule=rule_flow_vol_phase)

    def _flow_vol(self):

        def rule_flow_vol(b):
            return sum(b.flow_vol_phase[p] for p in self.params.phase_list)
        self.flow_vol = Expression(rule=rule_flow_vol)

    def _conc_mol_phase_comp(self):
        self.conc_mol_phase_comp = Var(
            self.params.phase_list,
            self.params.component_list,
            initialize=10,
            bounds=(1e-8, None),
            units=pyunits.mol * pyunits.m ** -3,
            doc="Molar concentration")

        def rule_conc_mol_phase_comp(b, j):
            return (b.conc_mol_phase_comp['Liq', j] * b.params.mw_comp[j] ==
                    b.conc_mass_phase_comp['Liq', j])
        self.eq_conc_mol_phase_comp = Constraint(self.params.component_list, rule=rule_conc_mol_phase_comp)

    def _conc_mass_phase_comp(self):
        self.conc_mass_phase_comp = Var(
            self.params.phase_list,
            self.params.component_list,
            initialize=10,
            bounds=(1e-8, 2e3),
            units=pyunits.kg * pyunits.m ** -3,
            doc="Mass concentration")

        def rule_conc_mass_phase_comp(b, j):
            return (b.conc_mass_phase_comp['Liq', j] ==
                    b.dens_mass_phase['Liq'] * b.mass_frac_phase_comp['Liq', j])
        self.eq_conc_mass_phase_comp = Constraint(self.params.component_list, rule=rule_conc_mass_phase_comp)

    def _flow_mass_phase_comp(self):
        self.flow_mass_phase_comp = Var(
            self.params.phase_list,
            self.params.component_list,
            initialize=100,
            bounds=(1e-8, None),
            units=pyunits.kg / pyunits.s,
            doc="Component Mass flowrate")

        def rule_flow_mass_phase_comp(b, j):
            return (b.flow_mass_phase_comp['Liq', j] ==
                    b.flow_mol_phase_comp['Liq', j] * b.params.mw_comp[j])
        self.eq_flow_mass_phase_comp = Constraint(self.params.component_list, rule=rule_flow_mass_phase_comp)

    def _mole_frac_phase_comp(self):
        self.mole_frac_phase_comp = Var(
            self.params.phase_list,
            self.params.component_list,
            initialize=0.1,
            bounds=(1e-8, 1.001),
            units=pyunits.dimensionless,
            doc="Mole fraction")

        def rule_mole_frac_phase_comp(b, j):
            return (b.mole_frac_phase_comp['Liq', j] == b.flow_mol_phase_comp['Liq', j] /
                    sum(b.flow_mol_phase_comp['Liq', j] for j in b.params.component_list))
        self.eq_mole_frac_phase_comp = Constraint(self.params.component_list, rule=rule_mole_frac_phase_comp)

    def _molality_comp(self):
        self.molality_comp = Var(
            self.params.solute_set,
            initialize=1,
            bounds=(1e-8, 10),
            units=pyunits.mole / pyunits.kg,
            doc="Molality")

        def rule_molality_comp(b, j):
            return (b.molality_comp[j] ==
                    b.flow_mol_phase_comp['Liq', j]
                    / b.flow_mol_phase_comp['Liq', 'H2O']
                    / b.params.mw_comp['H2O'])

        self.eq_molality_comp = Constraint(self.params.solute_set, rule=rule_molality_comp)

    def _radius_stokes_comp(self):
        add_object_reference(self, "radius_stokes_comp", self.params.radius_stokes_comp)

    def _diffus_phase_comp(self):
        add_object_reference(self, "diffus_phase_comp", self.params.diffus_phase_comp)

    def _visc_d_phase(self):
        add_object_reference(self, "visc_d_phase", self.params.visc_d_phase)

    def _mw_comp(self):
        add_object_reference(self, "mw_comp", self.params.mw_comp)

    def _charge_comp(self):
        add_object_reference(self, "charge_comp", self.params.charge_comp)

    def _dielectric_constant(self):
        add_object_reference(self, "dielectric_constant", self.params.dielectric_constant)

    def _act_coeff_phase_comp(self):
        self.act_coeff_phase_comp = Var(
            self.phase_list,
            self.params.solute_set,
            initialize=0.7,
            domain=NonNegativeReals,
            bounds=(1e-3, 1.001),
            units=pyunits.dimensionless,
            doc="activity coefficient of component")

        def rule_act_coeff_phase_comp(b, j):
            if b.params.config.activity_coefficient_model == ActivityCoefficientModel.ideal:
                return b.act_coeff_phase_comp['Liq', j] == 1
            elif b.params.config.activity_coefficient_model == ActivityCoefficientModel.davies:
                I = b.ionic_strength
                return (log(b.act_coeff_phase_comp['Liq', j]) == -b.debye_huckel_constant
                        * b.charge_comp[j] ** 2
                        * (I**0.5
                        / (1 * pyunits.mole ** 0.5 / pyunits.kg ** 0.5
                           + I ** 0.5)
                           - b.params.debye_huckel_b * I))
        self.eq_act_coeff_phase_comp = Constraint(self.params.solute_set,
                                                  rule=rule_act_coeff_phase_comp)
    #TODO: note- assuming molal ionic strength goes into Debye Huckel relationship;
    # the MIT's DSPMDE paper indicates usage of molar concentration
    def _ionic_strength(self):
        self.ionic_strength = Var(
            initialize = 1,
            domain=NonNegativeReals,
            units=pyunits.mol/pyunits.kg,
            doc="Molal ionic strength")

        def rule_ionic_strength(b):
            return b.ionic_strength == 0.5 * sum(b.charge_comp[j]**2 * b.molality_comp[j]
                                                 for j in self.params.solute_set)
        self.eq_ionic_strength = Constraint(rule=rule_ionic_strength)

    def _debye_huckel_constant(self):
        self.debye_huckel_constant = Var(
            initialize = 1,
            domain=NonNegativeReals,
            units=pyunits.dimensionless,
            # TODO: units are technically (kg/mol)**0.5, but Debye Huckel equation
            #  is empirical and units don't seem to cancel as typical. leaving as dimensionless for now
            doc="Temperature-dependent Debye Huckel constant A")

        def rule_debye_huckel_constant(b):
            return (b.debye_huckel_constant == ((2 * Constants.pi * Constants.avogadro_number) ** 0.5
                                                / log(10)) * (Constants.elemental_charge ** 2
                                                             /(4 * Constants.pi
                                                               * Constants.vacuum_electric_permittivity
                                                               * b.params.dielectric_constant
                                                               * Constants.boltzmann_constant
                                                               * b.temperature)) ** (3/2)
                    * (pyunits.coulomb ** 3
                    * pyunits.m ** 1.5
                    / pyunits.farad ** 1.5
                    / pyunits.J ** 1.5
                    / pyunits.mol ** 0.5)**-1)
        self.eq_debye_huckel_constant = Constraint(rule=rule_debye_huckel_constant)

    #TODO: change osmotic pressure calc
    def _pressure_osm(self):
        self.pressure_osm = Var(
            initialize=1e6,
            bounds=(5e2, 5e7),
            units=pyunits.Pa,
            doc="van't Hoff Osmotic pressure")

        def rule_pressure_osm(b):
            return (b.pressure_osm ==
                    sum(b.conc_mol_phase_comp['Liq', j] for j in self.params.solute_set)
                    * Constants.gas_constant * b.temperature)
        self.eq_pressure_osm = Constraint(rule=rule_pressure_osm)

    # -----------------------------------------------------------------------------
    # General Methods
    # NOTE: For scaling in the control volume to work properly, these methods must
    # return a pyomo Var or Expression

    def get_material_flow_terms(self, p, j):
        """Create material flow terms for control volume."""
        return self.flow_mol_phase_comp[p, j]

    # TODO: add enthalpy terms later
    # def get_enthalpy_flow_terms(self, p):
    #     """Create enthalpy flow terms."""
    #     return self.enth_flow

    # TODO: make property package compatible with dynamics
    # def get_material_density_terms(self, p, j):
    #     """Create material density terms."""

    # def get_enthalpy_density_terms(self, p):
    #     """Create enthalpy density terms."""

    def default_material_balance_type(self):
        return MaterialBalanceType.componentTotal

    # TODO: augment model with energybalance later
    # def default_energy_balance_type(self):
    #     return EnergyBalanceType.enthalpyTotal

    def get_material_flow_basis(self):
        return MaterialFlowBasis.molar

    def define_state_vars(self):
        """Define state vars."""
        return {"flow_mol_phase_comp": self.flow_mol_phase_comp,
                "temperature": self.temperature,
                "pressure": self.pressure}

    def assert_electroneutrality(self, tol=None, tee=True, defined_state=True, adjust_by_ion=None,
                                 get_property=None, solve=True):
        if tol is None:
            tol = 1e-8
        if not defined_state and get_property is not None:
            raise ValueError(f'Set defined_state to true if get_property = {get_property}')
        if adjust_by_ion is not None:
            if adjust_by_ion in self.params.solute_set:
                self.charge_balance = Constraint(expr=sum(self.charge_comp[j] * self.conc_mol_phase_comp['Liq', j]
                                                 for j in self.params.solute_set) == 0)
            else:
                raise ValueError('adjust_by_ion must be set to the name of an ion in the list of solutes.')
        if defined_state:
            for j in self.params.solute_set:
                if not self.flow_mol_phase_comp['Liq', j].is_fixed() and adjust_by_ion != j:
                    raise AssertionError(
                        f"{self.flow_mol_phase_comp['Liq', j]} was not fixed. Fix flow_mol_phase_comp for each solute"
                        f" to check that electroneutrality is satisfied.")
                if adjust_by_ion == j and self.flow_mol_phase_comp['Liq', j].is_fixed():
                    self.flow_mol_phase_comp['Liq', j].unfix()
        else:
            for j in self.params.solute_set:
                if self.flow_mol_phase_comp['Liq', j].is_fixed():
                    raise AssertionError(
                        f"{self.flow_mol_phase_comp['Liq', j]} was fixed. Either set defined_state=True or unfix "
                        f"flow_mol_phase_comp for each solute to check that electroneutrality is satisfied.")

        # touch this var since it is required for this method
        self.conc_mol_phase_comp

        if solve:
            solve = get_solver()
            results = solve.solve(self)
            if check_optimal_termination(results):
                val = value(sum(self.charge_comp[j] * self.conc_mol_phase_comp['Liq', j]
                            for j in self.params.solute_set))
            else:
                if adjust_by_ion is not None:
                    del self.charge_balance
                raise ValueError('The stateblock failed to solve while computing concentrations to check the charge balance.')
        else:
            val = value(sum(self.charge_comp[j] * self.conc_mol_phase_comp['Liq', j]
                            for j in self.params.solute_set))
        if abs(val) <= tol:
            if adjust_by_ion is not None:
                del self.charge_balance
                ion_adjusted = self.flow_mol_phase_comp['Liq', adjust_by_ion].value
                if defined_state:
                    self.flow_mol_phase_comp['Liq', adjust_by_ion].fix(ion_adjusted)
                    # touch on-demand property desired
                    if get_property is not None:
                        if isinstance(get_property, str):
                            getattr(self, get_property)
                        elif isinstance(get_property, (list, tuple)):
                            for i in get_property:
                                getattr(self, i)
                        else:
                            raise TypeError("get_property must be a string or list/tuple of strings.")
                        res_with_prop = solve.solve(self)
                        if not check_optimal_termination(res_with_prop):
                            raise ValueError(f'The stateblock failed to solve while solving with on-demand property'
                                             f' {get_property}.')
                    msg = f"{adjust_by_ion} was adjusted and flow_mol_phase_comp['Liq',{adjust_by_ion}] was fixed " \
                          f"to {ion_adjusted}."
                else:
                    msg = f"{adjust_by_ion} was adjusted and the value computed for flow_mol_phase_comp['Liq',{adjust_by_ion}]" \
                          f" is {ion_adjusted}."

            else:
                msg = ""
            if tee:
                return print(f'{msg} Electroneutrality satisfied for {self}. Balance Result = {val}')

        else:
            raise AssertionError(f"Electroneutrality condition violated in {self}. Ion concentrations should be adjusted to bring "
                                 f"the result of {val:.2E} closer towards 0.")

    # -----------------------------------------------------------------------------
    # Scaling methods
    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        # setting scaling factors for variables

        # default scaling factors have already been set with
        # idaes.core.property_base.calculate_scaling_factors()
        # for the following variables: pressure,
        # temperature, dens_mass, visc_d_phase, diffus_phase_comp

        # these variables should have user input
        if iscale.get_scaling_factor(self.flow_mol_phase_comp['Liq', 'H2O']) is None:
            sf = iscale.get_scaling_factor(self.flow_mol_phase_comp['Liq', 'H2O'], default=1, warning=True)
            iscale.set_scaling_factor(self.flow_mol_phase_comp['Liq', 'H2O'], sf)

        for j in self.params.solute_set:
            if iscale.get_scaling_factor(self.flow_mol_phase_comp['Liq', j]) is None:
                sf = iscale.get_scaling_factor(self.flow_mol_phase_comp['Liq', j], default=1, warning=True)
                iscale.set_scaling_factor(self.flow_mol_phase_comp['Liq', j], sf)

        # scaling factors for parameters
        for j, v in self.mw_comp.items():
            if iscale.get_scaling_factor(v) is None:
                iscale.set_scaling_factor(self.mw_comp[j], 1e1)
        for ind, v in self.diffus_phase_comp.items():
            if iscale.get_scaling_factor(v) is None:
                iscale.set_scaling_factor(self.diffus_phase_comp[ind], 1e10)

        if self.is_property_constructed('dens_mass_solvent'):
            if iscale.get_scaling_factor(self.dens_mass_solvent) is None:
                iscale.set_scaling_factor(self.dens_mass_solvent, 1e-2)

        for p, v in self.dens_mass_phase.items():
            if iscale.get_scaling_factor(v) is None:
                iscale.set_scaling_factor(self.dens_mass_phase[p], 1e-2)
        for p, v in self.visc_d_phase.items():
            if iscale.get_scaling_factor(v) is None:
                iscale.set_scaling_factor(self.visc_d_phase[p], 1e3)

        if self.is_property_constructed('mole_frac_phase_comp'):
            for j in self.params.component_list:
                if iscale.get_scaling_factor(self.mole_frac_phase_comp['Liq', j]) is None:
                    if j == 'H2O':
                        iscale.set_scaling_factor(self.mole_frac_phase_comp['Liq', j], 1)
                    else:
                        sf = (iscale.get_scaling_factor(self.flow_mol_phase_comp['Liq', j])
                              / iscale.get_scaling_factor(self.flow_mol_phase_comp['Liq', 'H2O']))
                        iscale.set_scaling_factor(self.mole_frac_phase_comp['Liq', j], sf)

        if self.is_property_constructed('flow_mass_phase_comp'):
            for j in self.params.component_list:
                if iscale.get_scaling_factor(self.flow_mass_phase_comp['Liq', j]) is None:
                    sf = iscale.get_scaling_factor(self.flow_mol_phase_comp['Liq', j], default=1)
                    sf *= iscale.get_scaling_factor(self.mw_comp[j])
                    iscale.set_scaling_factor(self.flow_mass_phase_comp['Liq', j], sf)

        if self.is_property_constructed('mass_frac_phase_comp'):
            for j in self.params.component_list:
                comp = self.params.get_component(j)
                if iscale.get_scaling_factor(self.mass_frac_phase_comp['Liq', j]) is None:
                    if comp.is_solute():
                        sf = (iscale.get_scaling_factor(self.flow_mass_phase_comp['Liq', j], default=1)
                              / iscale.get_scaling_factor(self.flow_mass_phase_comp['Liq', 'H2O'], default=1))
                        iscale.set_scaling_factor(self.mass_frac_phase_comp['Liq', j], sf)
                    else:
                        iscale.set_scaling_factor(self.mass_frac_phase_comp['Liq', j], 100)

        if self.is_property_constructed('conc_mass_phase_comp'):
            for j in self.params.component_list:
                sf_dens = iscale.get_scaling_factor(self.dens_mass_phase['Liq'])
                if iscale.get_scaling_factor(self.conc_mass_phase_comp['Liq', j]) is None:
                    if j == 'H2O':
                        # solvents typically have a mass fraction between 0.5-1
                        iscale.set_scaling_factor(self.conc_mass_phase_comp['Liq', j], sf_dens)
                    else:
                        iscale.set_scaling_factor(
                            self.conc_mass_phase_comp['Liq', j],
                            sf_dens * iscale.get_scaling_factor(self.mass_frac_phase_comp['Liq', j],default=1,warning=True))

        if self.is_property_constructed('conc_mol_phase_comp'):
            for j in self.params.component_list:
                if iscale.get_scaling_factor(self.conc_mol_phase_comp['Liq', j]) is None:
                    sf = (iscale.get_scaling_factor(self.conc_mass_phase_comp['Liq', j])
                          / iscale.get_scaling_factor(self.mw_comp[j]))
                    iscale.set_scaling_factor(self.conc_mol_phase_comp['Liq', j], sf)

        # these variables do not typically require user input,
        # will not override if the user does provide the scaling factor
        if self.is_property_constructed('pressure_osm'):
            if iscale.get_scaling_factor(self.pressure_osm) is None:
                sf = iscale.get_scaling_factor(self.pressure)
                iscale.set_scaling_factor(self.pressure_osm, sf)

        if self.is_property_constructed('flow_vol_phase'):
            sf = (iscale.get_scaling_factor(self.flow_mol_phase_comp['Liq', 'H2O'], default=1)
                  * iscale.get_scaling_factor(self.mw_comp[j])
                  / iscale.get_scaling_factor(self.dens_mass_phase['Liq']))
            iscale.set_scaling_factor(self.flow_vol_phase, sf)

        if self.is_property_constructed('flow_vol'):
            sf = iscale.get_scaling_factor(self.flow_vol_phase)
            iscale.set_scaling_factor(self.flow_vol, sf)

        if self.is_property_constructed('molality_comp'):
            for j in self.params.solute_set:
                if iscale.get_scaling_factor(self.molality_comp[j]) is None:
                    sf = (iscale.get_scaling_factor(self.flow_mol_phase_comp['Liq', j])
                          / iscale.get_scaling_factor(self.flow_mol_phase_comp['Liq', 'H2O'])
                          / iscale.get_scaling_factor(self.mw_comp[j]))
                    iscale.set_scaling_factor(self.molality_comp[j], sf)

        if self.is_property_constructed('act_coeff_phase_comp'):
            for j in self.params.solute_set:
                if iscale.get_scaling_factor(self.act_coeff_phase_comp['Liq', j]) is None:
                    iscale.set_scaling_factor(self.act_coeff_phase_comp['Liq', j], 1)

        if self.is_property_constructed('debye_huckel_constant'):
            if iscale.get_scaling_factor(self.debye_huckel_constant) is None:
                iscale.set_scaling_factor(self.debye_huckel_constant, 10)

        if self.is_property_constructed('ionic_strength'):
            if iscale.get_scaling_factor(self.ionic_strength) is None:
                iscale.set_scaling_factor(self.ionic_strength, 1)

        # transforming constraints
        # property relationships with no index, simple constraint
        for v_str in ('pressure_osm', 'dens_mass_solvent'):
            if self.is_property_constructed(v_str):
                v = getattr(self, v_str)
                sf = iscale.get_scaling_factor(v, default=1, warning=True)
                c = getattr(self, 'eq_' + v_str)
                iscale.constraint_scaling_transform(c, sf)

        # # property relationships with phase index, but simple constraint
        for v_str in ('flow_vol_phase', 'dens_mass_phase'):
            if self.is_property_constructed(v_str):
                v = getattr(self, v_str)
                sf = iscale.get_scaling_factor(v['Liq'], default=1, warning=True)
                c = getattr(self, 'eq_' + v_str)
                iscale.constraint_scaling_transform(c, sf)

        # property relationship indexed by component
        v_str_lst_comp = ['molality_comp']
        for v_str in v_str_lst_comp:
            if self.is_property_constructed(v_str):
                v_comp = getattr(self, v_str)
                c_comp = getattr(self, 'eq_' + v_str)
                for j, c in c_comp.items():
                    sf = iscale.get_scaling_factor(v_comp[j], default=1, warning=True)
                    iscale.constraint_scaling_transform(c, sf)

        # property relationships indexed by component and phase
        v_str_lst_phase_comp = ['mass_frac_phase_comp', 'conc_mass_phase_comp', 'flow_mass_phase_comp',
                                'mole_frac_phase_comp', 'conc_mol_phase_comp', 'act_coeff_phase_comp']
        for v_str in v_str_lst_phase_comp:
            if self.is_property_constructed(v_str):
                v_comp = getattr(self, v_str)
                c_comp = getattr(self, 'eq_' + v_str)
                for j, c in c_comp.items():
                    sf = iscale.get_scaling_factor(v_comp['Liq', j], default=1, warning=True)
                    iscale.constraint_scaling_transform(c, sf)

        if self.is_property_constructed('debye_huckel_constant'):
            iscale.constraint_scaling_transform(self.eq_debye_huckel_constant, 10)

        if self.is_property_constructed('ionic_strength'):
            iscale.constraint_scaling_transform(self.eq_ionic_strength, 1)



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
This module contains a zero-order representation of a coagulation/flocculation unit.
operation.
"""

from pyomo.environ import Constraint, units as pyunits, Var, Expression
from idaes.core import declare_process_block_class

from watertap.core import build_pt, constant_intensity, ZeroOrderBaseData

# Some more information about this module
__author__ = "Adam Atia"


@declare_process_block_class("CoagulationFlocculationZO")
class CoagulationFlocculationZOData(ZeroOrderBaseData):
    """
    Zero-Order model for a Coagulation/Flocculation unit operation.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "coag_and_floc"

        build_pt(self)

        self.alum_dose = Var(self.flowsheet().time,
                             units=pyunits.mg / pyunits.L,
                             bounds=(0, None),
                             doc="Dosing rate of alum")

        self.polymer_dose = Var(self.flowsheet().time,
                             units=pyunits.mg / pyunits.L,
                             bounds=(0, None),
                             doc="Dosing rate of polymer")

        self.anion_to_cation_polymer_ratio = Var(
            self.flowsheet().time,
            bounds=(0, None),
            units=pyunits.dimensionless,
            doc="Ratio of anionic to cationic polymer in dosage")

        self.anionic_polymer_dose = Var(self.flowsheet().config.time,
                                        bounds=(0, None),
                                        units=pyunits.mg/pyunits.L,
                                        doc="Dosing rate of anionic polymer")

        self.cationic_polymer_dose = Var(self.flowsheet().config.time,
                                        bounds=(0, None),
                                        units=pyunits.mg / pyunits.L,
                                        doc="Dosing rate of cationic polymer")

        self.chemical_flow_mass = Var(
            self.flowsheet().time,
            ["alum", "polymer"],
            units=pyunits.kg/pyunits.s,
            bounds=(0, None),
            doc="Mass flow rate of chemical solution")

        self.rapid_mix_retention_time = Var(self.flowsheet().config.time,
                                            units=pyunits.seconds,
                                            doc="Rapid Mix Retention Time")

        self.floc_retention_time = Var(self.flowsheet().config.time,
                                       units=pyunits.minutes,
                                       doc="Floc Retention Time")

        self.rapid_mix_basin_vol = Var(units=pyunits.m**3,
                                       doc="Rapid Mix Basin Volume")

        self.floc_basin_vol = Var(units=pyunits.m**3,
                                  doc="Floc Basin Volume")

        self.num_rapid_mixers = Var(units=pyunits.dimensionless,
                                    doc="Number of Rapid Mixers")

        self.num_floc_mixers = Var(units=pyunits.dimensionless,
                                    doc="Number of Floc Mixers")

        self.num_rapid_mix_processes = Var(units=pyunits.dimensionless,
                                           doc="Number of Rapid Mix Processes")

        self.num_floc_processes = Var(units=pyunits.dimensionless,
                                       doc="Number of Floc Processes")

        self.num_coag_processes = Var(units=pyunits.dimensionless,
                                       doc="Number of Coagulation Processes")

        self.num_floc_injection_processes = Var(units=pyunits.dimensionless,
                                                doc="Number of Floc Injection Processes")

        self.velocity_gradient_rapid_mix = Var(self.flowsheet().config.time,
                               units=pyunits.s**-1,
                               doc="Rapid Mix Velocity Gradient")

        self.velocity_gradient_floc = Var(self.flowsheet().config.time,
                          units=pyunits.s**-1,
                          doc="Floc Velocity Gradient")

        self.power_rapid_mix = Var(self.flowsheet().config.time,
                                   units=pyunits.kW,
                                   doc="Rapid Mix Power Consumption")

        self.power_floc = Var(self.flowsheet().config.time,
                              units=pyunits.kW,
                              doc="Floc Power Consumption")

        self.electricity = Var(self.flowsheet().config.time,
                               units=pyunits.kW,
                               doc="Total Power Consumption")

        self._fixed_perf_vars.append(self.alum_dose)
        self._fixed_perf_vars.append(self.polymer_dose)
        self._fixed_perf_vars.append(self.anion_to_cation_polymer_ratio)
        self._fixed_perf_vars.append(self.rapid_mix_retention_time)
        self._fixed_perf_vars.append(self.floc_retention_time)
        self._fixed_perf_vars.append(self.num_floc_injection_processes)
        self._fixed_perf_vars.append(self.num_floc_processes)
        self._fixed_perf_vars.append(self.num_rapid_mixers)
        self._fixed_perf_vars.append(self.num_coag_processes)
        self._fixed_perf_vars.append(self.num_floc_mixers)
        self._fixed_perf_vars.append(self.num_rapid_mix_processes)
        self._fixed_perf_vars.append(self.velocity_gradient_rapid_mix)
        self._fixed_perf_vars.append(self.velocity_gradient_floc)

        self._perf_var_dict["Alum Dosage (mg/L)"] = self.alum_dose
        self._perf_var_dict["Polymer Dosage (mg/L)"] = self.polymer_dose
        self._perf_var_dict["Alum Flow (kg/s)"] = self.chemical_flow_mass[0, "alum"]
        self._perf_var_dict["Polymer Flow (kg/s)"] = self.chemical_flow_mass[0, "polymer"]
        self._perf_var_dict["Rapid Mix Basin Volume (m^3)"] = self.rapid_mix_basin_vol
        self._perf_var_dict["Floc Basin Volume (m^3)"] = self.floc_basin_vol
        self._perf_var_dict["Rapid Mix Retention Time (s)"] = self.rapid_mix_retention_time
        self._perf_var_dict["Floc Retention Time (min)"] = self.floc_retention_time
        self._perf_var_dict["Rapid Mix Velocity Gradient (1/s)"] = self.velocity_gradient_rapid_mix
        self._perf_var_dict["Floc Velocity Gradient (1/s)"] = self.velocity_gradient_floc
        self._perf_var_dict["Rapid Mix Power (kW)"] = self.power_rapid_mix
        self._perf_var_dict["Floc Power (kW)"] = self.power_floc
        self._perf_var_dict["Total Power Consumption (kW)"] = self.electricity

        def rule_rapid_mix_basin_vol(blk):
            return (blk.rapid_mix_basin_vol ==
                    blk.properties[0].flow_vol * blk.rapid_mix_retention_time[0])
        self.rapid_mix_basin_vol_constraint = Constraint(rule=rule_rapid_mix_basin_vol)

        def rule_floc_basin_vol(blk):
            return (blk.floc_basin_vol ==
                    pyunits.convert(blk.properties[0].flow_vol * blk.floc_retention_time[0],
                                    to_units=pyunits.m**3))
        self.floc_basin_vol_constraint = Constraint(rule=rule_floc_basin_vol)

        def rule_chem_flow(blk, t, j):
            if j == 'alum':
                chemical_dosage = blk.alum_dose[t]
            elif j == 'polymer':
                chemical_dosage = blk.polymer_dose[t]
            return (blk.chemical_flow_mass[t, j] ==
                    pyunits.convert(chemical_dosage * blk.properties[t].flow_vol,
                                    to_units=pyunits.kg/pyunits.s))
        self.chemical_flow_constraint = Constraint(self.flowsheet().time,
                                                   ["alum", "polymer"],
                                                   rule=rule_chem_flow)

        def rule_anionic_polymer_dose(blk, t):
            return (blk.anionic_polymer_dose[t] ==
                    blk.anion_to_cation_polymer_ratio[t]
                    * blk.polymer_dose[t]
                    / (blk.anion_to_cation_polymer_ratio[t] + 1))
        self.anionic_polymer_dose_constraint = Constraint(self.flowsheet().config.time,
                                                          rule=rule_anionic_polymer_dose)

        def rule_cationic_polymer_dose(blk, t):
            return (blk.cationic_polymer_dose[t] == blk.polymer_dose[t]
                    / (blk.anion_to_cation_polymer_ratio[t] + 1))
        self.cationic_polymer_dose_constraint = Constraint(self.flowsheet().config.time,
                                                           rule=rule_cationic_polymer_dose)

        # TODO: WT3 doesn't include pump electricity. Consider adding after case study validation
        # pump_electricity(self, self.chemical_flow_vol)

        @self.Constraint(self.flowsheet().time,
                         doc='Constraint for rapid mix power consumption')
        def rule_power_rapid_mix(b, t):
            return (b.power_rapid_mix[t] ==
                    pyunits.convert(b.num_rapid_mixers
                    * b.properties[t].visc_d
                    * b.rapid_mix_basin_vol
                    * b.velocity_gradient_rapid_mix[t] ** 2, to_units=pyunits.kW))

        @self.Constraint(self.flowsheet().time,
                         doc='Constraint for floc power consumption')
        def rule_power_floc(b, t):
            return (b.power_floc[t] ==
                    pyunits.convert(b.num_floc_mixers
                    * b.properties[t].visc_d
                    * b.floc_basin_vol
                    * b.velocity_gradient_floc[t] ** 2, to_units=pyunits.kW))

        @self.Constraint(self.flowsheet().time,
                         doc='Total power consumption')
        def electricity_constraint(b, t):
            return b.electricity[t] == b.power_floc[t] + b.power_rapid_mix[t]



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
This module contains a zero-order representation of an anaerobic digestion oxidation unit.
operation.
"""

from pyomo.environ import Constraint, units as pyunits, Var
from idaes.core import declare_process_block_class

from watertap.core import build_sido, constant_intensity, ZeroOrderBaseData

# Some more information about this module
__author__ = "Chenyu Wang"


@declare_process_block_class("AnaerobicDigestionOxidationZO")
class AnaerobicDigestionOxidationZOData(ZeroOrderBaseData):
    """
    Zero-Order model for an Anaerobic Digestion Oxidation unit operation.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "anaerobic_digestion_oxidation"

        build_sido(self)
        constant_intensity(self)


###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
This module contains a zero-order representation of a Ozone-AOP unit.
operation.
"""

from pyomo.environ import units as pyunits, Var
from idaes.core import declare_process_block_class
from watertap.unit_models.zero_order.ozone_zo import OzoneZOData

# Some more information about this module
__author__ = "Kurban Sitterley"


@declare_process_block_class("OzoneAOPZO")
class OzoneAOPZOData(OzoneZOData):
    """
    Zero-Order model for a Ozone-AOP unit operation.
    """

    def build(self):
        super().build()

        self._tech_type = "ozone_aop"

        self.oxidant_dose = Var(self.flowsheet().time,
                                units=pyunits.mg/pyunits.L,
                                doc="Oxidant dosage")

        self.chemical_flow_mass = Var(
            self.flowsheet().time,
            units=pyunits.kg/pyunits.s,
            bounds=(0, None),
            doc="Mass flow rate of oxidant solution")

        self.ozone_toc_ratio = Var(self.flowsheet().time,
                                   units=pyunits.dimensionless,
                                   doc="Ratio of ozone to total organic carbon")

        self.oxidant_ozone_ratio = Var(self.flowsheet().time,
                                                 units=pyunits.dimensionless,
                                                 doc="Ratio of oxidant to ozone")

        self._fixed_perf_vars.append(self.oxidant_ozone_ratio)

        @self.Constraint(self.flowsheet().time,
                         doc="Ozone/TOC ratio constraint")
        def ozone_toc_ratio_constraint(b, t):
            return (b.ozone_toc_ratio[t] == 1
                    + pyunits.convert(b.concentration_time[t] / b.contact_time[t]
                    / b.properties_in[t].conc_mass_comp['toc'],
                                      to_units=pyunits.dimensionless))

        @self.Constraint(self.flowsheet().time,
                         doc="Oxidant dose constraint")
        def oxidant_dose_constraint(b,t):
            return (b.oxidant_dose[t] == pyunits.convert(
                b.oxidant_ozone_ratio[t]
                * b.ozone_toc_ratio[t]
                * b.properties_in[t].conc_mass_comp["toc"],
                to_units=pyunits.mg/pyunits.L))

        @self.Constraint(self.flowsheet().time,
                         doc="Oxidant mass flow constraint")
        def chemical_flow_mass_constraint(b, t):
            return (b.chemical_flow_mass[t] ==
                    pyunits.convert(
                        b.oxidant_dose[t]*b.properties_in[t].flow_vol,
                        to_units=pyunits.kg/pyunits.s))

        self._perf_var_dict["Oxidant Dosage (mg/L)"] = self.oxidant_dose
        self._perf_var_dict["Oxidant Flow (kg/s)"] = self.chemical_flow_mass
        self._perf_var_dict["Oxidant/Ozone Ratio"] = self.oxidant_ozone_ratio
        self._perf_var_dict["Ozone/TOC Ratio"] = self.ozone_toc_ratio




###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
This module contains a zero-order representation of an evaporation pond unit
model.
"""

from pyomo.environ import units as pyunits, Var, Constraint, Reference, value, sqrt
from pyomo.common.config import ConfigValue, In
from idaes.core import declare_process_block_class
from watertap.core import build_sido, constant_intensity, ZeroOrderBaseData

# Some more information about this module
__author__ = "Kurban Sitterley"


@declare_process_block_class("EvaporationPondZO")
class EvaporationPondZOData(ZeroOrderBaseData):
    """
    Zero-Order model for a evaporation pond unit.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        build_sido(self)

        self._tech_type = "evaporation_pond"

        self.air_temperature = Var(self.flowsheet().time,
                    initialize=298,
                    units=pyunits.kelvin,
                    doc="Air temperature")

        self.solar_radiation = Var(self.flowsheet().time,
                    units=pyunits.mJ/pyunits.m**2,
                    doc="Daily solar radiation incident")

        self.dike_height = Var(self.flowsheet().time,
                    units=pyunits.ft,
                    doc="Pond dike height")  

        self.evaporation_rate_adj_factor = Var(self.flowsheet().time,
                    units=pyunits.dimensionless,
                    doc="Factor to adjust evaporation rate of pure water")

        self.evap_rate_calc_a_parameter = Var(self.flowsheet().time,
                    units=pyunits.mm/pyunits.d,
                    doc="Evaporation rate calculation parameter A")  

        self.evap_rate_calc_b_parameter = Var(self.flowsheet().time,
                    units=pyunits.m**2/pyunits.mJ,
                    doc="Evaporation rate calculation parameter B")  

        self.evap_rate_calc_c_parameter = Var(self.flowsheet().time,
                    units=pyunits.m**2/pyunits.mJ,
                    doc="Evaporation rate calculation parameter C")

        self.adj_area_calc_a_parameter = Var(self.flowsheet().time,
                    units=pyunits.acres,
                    doc="Adjusted area calculation parameter A")  

        self.adj_area_calc_b_parameter = Var(self.flowsheet().time,
                    units=pyunits.dimensionless,
                    doc="Adjusted area calculation parameter B") 

        self._fixed_perf_vars.append(self.air_temperature)
        self._fixed_perf_vars.append(self.solar_radiation)
        self._fixed_perf_vars.append(self.dike_height)
        self._fixed_perf_vars.append(self.evaporation_rate_adj_factor)
        self._fixed_perf_vars.append(self.evap_rate_calc_a_parameter)
        self._fixed_perf_vars.append(self.evap_rate_calc_b_parameter)
        self._fixed_perf_vars.append(self.evap_rate_calc_c_parameter) 
        self._fixed_perf_vars.append(self.adj_area_calc_a_parameter)
        self._fixed_perf_vars.append(self.adj_area_calc_b_parameter)  
            
        self.area = Var(self.flowsheet().time,
                        initialize=1, 
                        units=pyunits.acres,
                        bounds=(0, None),
                        doc="Pond area needed based on evaporation rate")

        self.adj_area = Var(self.flowsheet().time,
                                units=pyunits.acres,
                                doc="Adjusted pond area needed")  
                                            
        self.evaporation_rate_pure = Var(self.flowsheet().time,
                                units=pyunits.mm/pyunits.d,
                                doc="Calculated evaporation rate of pure water")

        self.evaporation_rate_salt = Var(self.flowsheet().time,
                                units=(pyunits.gallons/pyunits.minute/pyunits.acre),
                                doc="Pure water evaporation rate adjusted for salinity")

        @self.Constraint(self.flowsheet().time,
                    doc='Evaporation rate of pure water constraint')
        def evap_rate_pure_constraint(b, t):
            air_temperature_C = pyunits.convert_temp_K_to_C(value(b.air_temperature[t]))
            return (b.evaporation_rate_pure[t] ==
                    b.evap_rate_calc_a_parameter[t] * (b.evap_rate_calc_b_parameter[t] * 
                    air_temperature_C + b.evap_rate_calc_c_parameter[t]) 
                    * b.solar_radiation[t])

        @self.Constraint(self.flowsheet().time,
                    doc='Adjusted evaporation rate for salinity constraint')
        def evap_rate_salt_constraint(b, t):
            evap_rate_gal_min_acre = pyunits.convert(b.evaporation_rate_pure[t], 
                        to_units=(pyunits.gallons/pyunits.minute/pyunits.acre))
            return (b.evaporation_rate_salt[t] ==
                    evap_rate_gal_min_acre * b.evaporation_rate_adj_factor[t])

        @self.Constraint(self.flowsheet().time,
                    doc='Base area constraint')
        def area_constraint(b, t):
            q_out = pyunits.convert(self.properties_byproduct[t].flow_vol, 
            to_units=pyunits.gallon/pyunits.minute)
            return (q_out == b.evaporation_rate_salt[t] * b.area[t])

        @self.Constraint(self.flowsheet().time,
                    doc='Adjusted area constraint')
        def area_adj_constraint(b, t):
            area = b.area[t]/pyunits.acres
            dike_ht = b.dike_height[t]/pyunits.ft
            return (b.adj_area[t] == 
                b.adj_area_calc_a_parameter[t] * area * 
                (1 + (b.adj_area_calc_b_parameter[t] * dike_ht)/ area ** 0.5))

        self._perf_var_dict["Evaporation rate (mm/d)"] = self.evaporation_rate_pure
        self._perf_var_dict["Pond area (acres)"] = self.adj_area
        self._perf_var_dict["Pond dike height (ft)"] = self.dike_height
    


###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
This module contains a zero-order representation of an iron and manganese removal unit
"""

from pyomo.environ import Constraint, units as pyunits, Var, Param
from idaes.core import declare_process_block_class
from watertap.core import build_sido, constant_intensity, ZeroOrderBaseData

# Some more information about this module
__author__ = "Chenyu Wang"


@declare_process_block_class("IronManganeseRemovalZO")
class IronManganeseRemovalZOData(ZeroOrderBaseData):
    """
    Zero-Order model for an iron and manganese removal unit operation.
    """

    CONFIG = ZeroOrderBaseData.CONFIG()

    def build(self):
        super().build()

        self._tech_type = "iron_and_manganese_removal"

        build_sido(self)

        self.air_water_ratio = Var(self.flowsheet().time,
                                units=pyunits.dimensionless,
                                doc="Ratio of air to water")

        self.flow_basis = Var(self.flowsheet().time,
                                   units=pyunits.m**3/pyunits.hour,
                                   doc="Flow basis")

        self.air_flow_rate = Var(self.flowsheet().time,
                                   units=pyunits.m ** 3 / pyunits.hour,
                                   doc="Air flow rate")

        self.electricity_intensity_parameter = Var(units=pyunits.hp/(pyunits.m ** 3 / pyunits.hour),
                              doc="Constant in electricity intensity equation")

        self.filter_surf_area = Var(units=pyunits.m ** 2,
                                    doc="Dual media filter surface area")

        self.num_filter_units = Var(units=pyunits.dimensionless,
                                    doc="Number of dual media filter units")

        self.electricity = Var(self.flowsheet().config.time,
                                units=pyunits.kW,
                                bounds=(0, None),
                                doc="Power consumption of iron and manganese removal")

        self.electricity_intensity = Var(self.flowsheet().config.time,
                                         units=pyunits.kWh / pyunits.m ** 3,
                                         doc="Specific energy consumption with respect to feed flowrate")

        self._fixed_perf_vars.append(self.air_water_ratio)
        self._fixed_perf_vars.append(self.flow_basis)
        self._fixed_perf_vars.append(self.electricity_intensity_parameter)
        self._fixed_perf_vars.append(self.filter_surf_area)
        self._fixed_perf_vars.append(self.num_filter_units)

        @self.Constraint(self.flowsheet().config.time,
                         doc="Air flow rate constraint")
        def air_flow_rate_constraint(b, t):
            q_in = pyunits.convert(b.properties_in[t].flow_vol,
                                   to_units=pyunits.m**3/pyunits.hour)
            return (b.air_flow_rate[t] ==
                    b.air_water_ratio[t] * q_in)

        @self.Constraint(self.flowsheet().config.time,
                         doc="Electricity intensity constraint")
        def electricity_intensity_constraint(b, t):
            return b.electricity_intensity[t] == pyunits.convert(b.electricity_intensity_parameter
                                                                 * b.air_water_ratio[t],
                                                                 to_units=pyunits.kWh/pyunits.m**3)

        @self.Constraint(self.flowsheet().config.time,
                         doc="Power consumption constraint")
        def electricity_constraint(b, t):
            q_in = pyunits.convert(b.properties_in[t].flow_vol,
                                   to_units=pyunits.m**3/pyunits.hour)
            return (b.electricity[t] ==
                    b.electricity_intensity[t] * q_in)

        self._perf_var_dict["Power Consumption (kW)"] = self.electricity
        self._perf_var_dict["Electricity intensity per Inlet Flowrate  (kWh/m3)"] = \
            self.electricity_intensity



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Tests for zero-order cofermentation model
"""
import pytest

from io import StringIO
from pyomo.environ import (
    ConcreteModel, Constraint, value, Var, assert_optimal_termination, TransformationFactory,
    units as pyunits)
from pyomo.util.check_units import assert_units_consistent
from pyomo.network import Arc

import idaes.core.util.scaling as iscale
from idaes.core import FlowsheetBlock
from idaes.core.util import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester

from watertap.unit_models.zero_order import CofermentationZO, FeedZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock

solver = get_solver()

class TestCofermentationZO:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(
            default={"solute_list": ["cod"]})

        m.fs.unit = CofermentationZO(default={
            "property_package": m.fs.params,
            "database": m.db})
        m.fs.feed1 = FeedZO(default={
            "property_package": m.fs.params,
        })
        m.fs.feed2 = FeedZO(default={
            "property_package": m.fs.params,
        })

        m.fs.feed1.flow_vol[0].fix(value(pyunits.convert(32.1 * pyunits.L / pyunits.day,
                                                         to_units=pyunits.m ** 3 / pyunits.s)))
        m.fs.feed1.conc_mass_comp[0, 'cod'].fix(value(pyunits.convert(69781.93146 * pyunits.mg / pyunits.L,
                                                                      to_units=pyunits.kg / pyunits.m ** 3)))

        m.fs.feed2.flow_vol[0].fix(value(pyunits.convert(3.21 * pyunits.L / pyunits.day,
                                                         to_units=pyunits.m ** 3 / pyunits.s)))
        m.fs.feed2.conc_mass_comp[0, 'cod'].fix(value(pyunits.convert(1e4 * pyunits.mg / pyunits.L,
                                                                      to_units=pyunits.kg / pyunits.m ** 3)))

        m.fs.feed1_to_coferm = Arc(source=m.fs.feed1.outlet, destination=m.fs.unit.inlet1)
        m.fs.feed2_to_coferm = Arc(source=m.fs.feed2.outlet, destination=m.fs.unit.inlet2)

        TransformationFactory("network.expand_arcs").apply_to(m)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database is model.db

        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.energy_electric_flow_vol_inlet, Var)
        assert isinstance(model.fs.unit.electricity_consumption, Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("cofermentation")

        model.fs.unit.load_parameters_from_database(use_default_removal=True)

        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == \
            data["recovery_frac_mass_H2O"]["value"]

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            assert v.value == data["removal_frac_mass_solute"][j]["value"]

        assert model.fs.unit.energy_electric_flow_vol_inlet.fixed
        assert model.fs.unit.energy_electric_flow_vol_inlet.value == data[
            "energy_electric_flow_vol_inlet"]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert (pytest.approx(7.8892e-9, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].flow_vol))
        assert (pytest.approx( 0.68163, rel=1e-5) ==
                value(pyunits.convert(model.fs.unit.properties_treated[0].flow_mass_comp["cod"],
                                      to_units=pyunits.kg/pyunits.day)))
        assert (pytest.approx(0, abs=1e-5) ==
                value(model.fs.unit.electricity[0]))

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.solute_set:
            assert 1e-6 >= abs(value(
                ((1 - model.fs.unit.removal_frac_mass_solute[0, j]) *
                 (model.fs.unit.properties_in1[0].flow_mass_comp[j]
                 + model.fs.unit.properties_in2[0].flow_mass_comp[j]) -
                 model.fs.unit.properties_treated[0].flow_mass_comp[j])))
        assert 1e-6 >= abs((value(
            (model.fs.unit.properties_in1[0].flow_mass_comp["H2O"]
             + model.fs.unit.properties_in2[0].flow_mass_comp["H2O"])
             * model.fs.unit.recovery_frac_mass_H2O[0] -
             model.fs.unit.properties_treated[0].flow_mass_comp["H2O"])))

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()

        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                   : Value      : Fixed : Bounds
       Electricity Demand : 1.0000e-14 : False : (0, None)
    Electricity Intensity :     0.0000 :  True : (None, None)
     Solute Removal [cod] :    0.70000 :  True : (0, None)
           Water Recovery :     0.0000 :  True : (1e-08, 1.0000001)

------------------------------------------------------------------------------------
    Stream Table
                             Inlet 1    Inlet 2    Treated 
    Volumetric Flowrate    3.7153e-07 3.7153e-08 7.8892e-09
    Mass Concentration H2O     930.22     990.00 1.2675e-06
    Mass Concentration cod     69.782     10.000     1000.0
====================================================================================
"""

        assert output in stream.getvalue()


@pytest.mark.unit
def test_COD_not_in_solute_list():
    model = ConcreteModel()
    model.db = Database()

    model.fs = FlowsheetBlock(default={"dynamic": False})
    model.fs.params = WaterParameterBlock(default={"solute_list": ["foo"]})
    with pytest.raises(ValueError,
                       match="cod must be included in the solute list since"
                             " this unit model converts cod to nonbiodegradable_cod."):
        model.fs.unit = CofermentationZO(default={"property_package": model.fs.params,
                                                  "database": model.db})



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Tests for zero-order UV model
"""
import pytest
from io import StringIO

from pyomo.environ import (
    check_optimal_termination, ConcreteModel, Constraint, value, Var, Block)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.util import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.generic_models.costing import UnitModelCostingBlock

from watertap.unit_models.zero_order import UVZO, UVAOPZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock
from watertap.core.zero_order_costing import ZeroOrderCosting

solver = get_solver()


class TestUVZO_with_default_removal:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(
            default={"solute_list": ["viruses_enteric",
                                     "tss",
                                     "toc",
                                     "cryptosporidium",
                                     "total_coliforms_fecal_ecoli"]})

        m.fs.unit = UVZO(default={
            "property_package": m.fs.params,
            "database": m.db})


        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(10000)
        m.fs.unit.inlet.flow_mass_comp[0, "viruses_enteric"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "toc"].fix(2)
        m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(3)
        m.fs.unit.inlet.flow_mass_comp[0, "cryptosporidium"].fix(5)
        m.fs.unit.inlet.flow_mass_comp[0, "total_coliforms_fecal_ecoli"].fix(3)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database == model.db
        assert model.fs.unit._tech_type == 'uv'
        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.energy_electric_flow_vol_inlet, Var)
        assert isinstance(model.fs.unit.electricity_consumption, Constraint)
        assert isinstance(model.fs.unit.uv_reduced_equivalent_dose, Var)
        assert isinstance(model.fs.unit.uv_transmittance_in, Var)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("uv")
        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == 1
        model.fs.unit.load_parameters_from_database(use_default_removal=True)
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == 1

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            if j not in data["removal_frac_mass_solute"]:
                assert v.value == data["default_removal_frac_mass_solute"]["value"]
            else:
                assert v.value == data["removal_frac_mass_solute"][j]["value"]

        assert model.fs.unit.energy_electric_flow_vol_inlet.fixed
        assert model.fs.unit.energy_electric_flow_vol_inlet.value == data[
            "energy_electric_flow_vol_inlet"]["value"]
        assert model.fs.unit.uv_reduced_equivalent_dose[0].fixed
        assert model.fs.unit.uv_reduced_equivalent_dose[0].value == data[
            "uv_reduced_equivalent_dose"]["value"]
        assert model.fs.unit.uv_transmittance_in[0].fixed
        assert model.fs.unit.uv_transmittance_in[0].value == data[
            "uv_transmittance_in"]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert (pytest.approx(10.004926, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].flow_vol))
        assert (pytest.approx(0.1889939, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["toc"]))
        assert (pytest.approx(0.299852, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["tss"]))
        assert (pytest.approx(3605.04, rel=1e-5) ==
                value(model.fs.unit.electricity[0]))

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()

        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                                          : Value    : Fixed : Bounds
                              Electricity Demand :   3605.0 : False : (0, None)
                           Electricity Intensity :  0.10000 :  True : (None, None)
                Solute Removal [cryptosporidium] :  0.99999 :  True : (0, None)
                            Solute Removal [toc] : 0.054565 :  True : (0, None)
    Solute Removal [total_coliforms_fecal_ecoli] :  0.99999 :  True : (0, None)
                            Solute Removal [tss] :   0.0000 :  True : (0, None)
                Solute Removal [viruses_enteric] :  0.96540 :  True : (0, None)
          UV Reduced Equivalent Dosage (mJ/cm^2) :   100.00 :  True : (None, None)
                        UV Transmittance of Feed :  0.90000 :  True : (None, None)

------------------------------------------------------------------------------------
    Stream Table
                                                     Inlet    Treated 
    Volumetric Flowrate                              10.014     10.005
    Mass Concentration H2O                           998.60     999.51
    Mass Concentration viruses_enteric             0.099860  0.0034580
    Mass Concentration tss                          0.29958    0.29985
    Mass Concentration toc                          0.19972    0.18899
    Mass Concentration cryptosporidium              0.49930 5.4973e-06
    Mass Concentration total_coliforms_fecal_ecoli  0.29958 1.7991e-06
====================================================================================
"""

        assert output in stream.getvalue()


class TestUVZO_w_o_default_removal:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(default={"dynamic": False})
        m.fs.params = WaterParameterBlock(
            default={"solute_list": ["viruses_enteric",
                                     "toc",
                                     "cryptosporidium",
                                     "total_coliforms_fecal_ecoli"]})

        m.fs.unit = UVZO(default={
            "property_package": m.fs.params,
            "database": m.db})

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(10000)
        m.fs.unit.inlet.flow_mass_comp[0, "viruses_enteric"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "toc"].fix(2)
        m.fs.unit.inlet.flow_mass_comp[0, "cryptosporidium"].fix(5)
        m.fs.unit.inlet.flow_mass_comp[0, "total_coliforms_fecal_ecoli"].fix(3)
        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database == model.db
        assert model.fs.unit._tech_type == 'uv'
        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.energy_electric_flow_vol_inlet, Var)
        assert isinstance(model.fs.unit.electricity_consumption, Constraint)
        assert isinstance(model.fs.unit.uv_reduced_equivalent_dose, Var)
        assert isinstance(model.fs.unit.uv_transmittance_in, Var)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("uv")
        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == 1
        model.fs.unit.load_parameters_from_database()
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == 1

        for (t, j), v in model.fs.unit.removal_frac_mass_solute.items():
            assert v.fixed
            if j not in data["removal_frac_mass_solute"]:
                assert v.value == data["default_removal_frac_mass_solute"]["value"]
            else:
                assert v.value == data["removal_frac_mass_solute"][j]["value"]

        assert model.fs.unit.energy_electric_flow_vol_inlet.fixed
        assert model.fs.unit.energy_electric_flow_vol_inlet.value == data[
            "energy_electric_flow_vol_inlet"]["value"]
        assert model.fs.unit.uv_reduced_equivalent_dose[0].fixed
        assert model.fs.unit.uv_reduced_equivalent_dose[0].value == data[
            "uv_reduced_equivalent_dose"]["value"]
        assert model.fs.unit.uv_transmittance_in[0].fixed
        assert model.fs.unit.uv_transmittance_in[0].value == data[
            "uv_transmittance_in"]["value"]

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert (pytest.approx(10.00193, rel=1e-5) ==
                value(model.fs.unit.properties_treated[0].flow_vol))
        assert (pytest.approx(0.189051, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["toc"]))
        assert (pytest.approx(5.498941e-6, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["cryptosporidium"]))
        assert (pytest.approx(1.799653e-6, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["total_coliforms_fecal_ecoli"]))
        assert (pytest.approx(0.0034590, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["viruses_enteric"]))
        assert (pytest.approx(3603.96, rel=1e-5) ==
                value(model.fs.unit.electricity[0]))

    @pytest.mark.component
    def test_report(self, model):
        stream = StringIO()

        model.fs.unit.report(ostream=stream)

        output = """
====================================================================================
Unit : fs.unit                                                             Time: 0.0
------------------------------------------------------------------------------------
    Unit Performance

    Variables: 

    Key                                          : Value    : Fixed : Bounds
                              Electricity Demand :   3604.0 : False : (0, None)
                           Electricity Intensity :  0.10000 :  True : (None, None)
                Solute Removal [cryptosporidium] :  0.99999 :  True : (0, None)
                            Solute Removal [toc] : 0.054565 :  True : (0, None)
    Solute Removal [total_coliforms_fecal_ecoli] :  0.99999 :  True : (0, None)
                Solute Removal [viruses_enteric] :  0.96540 :  True : (0, None)
          UV Reduced Equivalent Dosage (mJ/cm^2) :   100.00 :  True : (None, None)
                        UV Transmittance of Feed :  0.90000 :  True : (None, None)

------------------------------------------------------------------------------------
    Stream Table
                                                     Inlet    Treated 
    Volumetric Flowrate                              10.011     10.002
    Mass Concentration H2O                           998.90     999.81
    Mass Concentration viruses_enteric             0.099890  0.0034590
    Mass Concentration toc                          0.19978    0.18905
    Mass Concentration cryptosporidium              0.49945 5.4989e-06
    Mass Concentration total_coliforms_fecal_ecoli  0.29967 1.7997e-06
====================================================================================
"""

        assert output in stream.getvalue()


def test_costing():
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.params = WaterParameterBlock(
        default={"solute_list": ["viruses_enteric",
                                 "toc",
                                 "cryptosporidium"]})

    m.fs.costing = ZeroOrderCosting()

    m.fs.unit1 = UVZO(default={
        "property_package": m.fs.params,
        "database": m.db})

    m.fs.unit1.inlet.flow_mass_comp[0, "H2O"].fix(10000)
    m.fs.unit1.inlet.flow_mass_comp[0, "viruses_enteric"].fix(1)
    m.fs.unit1.inlet.flow_mass_comp[0, "toc"].fix(2)
    m.fs.unit1.inlet.flow_mass_comp[0, "cryptosporidium"].fix(3)
    m.fs.unit1.load_parameters_from_database(use_default_removal=True)

    assert degrees_of_freedom(m.fs.unit1) == 0

    m.fs.unit1.costing = UnitModelCostingBlock(default={
        "flowsheet_costing_block": m.fs.costing})

    assert isinstance(m.fs.costing.uv, Block)
    assert isinstance(m.fs.costing.uv.uv_capital_a_parameter, Var)
    assert isinstance(m.fs.costing.uv.uv_capital_b_parameter, Var)
    assert isinstance(m.fs.costing.uv.uv_capital_c_parameter, Var)
    assert isinstance(m.fs.costing.uv.uv_capital_d_parameter, Var)
    assert isinstance(m.fs.unit1.costing.capital_cost, Var)
    assert isinstance(m.fs.unit1.costing.capital_cost_constraint,
                      Constraint)

    assert_units_consistent(m.fs)
    assert degrees_of_freedom(m.fs.unit1) == 0

    assert m.fs.unit1.electricity[0] in \
        m.fs.costing._registered_flows["electricity"]



###############################################################################
# WaterTAP Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#
###############################################################################
"""
Tests for general zero-order costing methods
"""
import pytest

from pyomo.environ import (Block,
                           check_optimal_termination,
                           ConcreteModel,
                           Constraint,
                           Expression,
                           Set,
                           units as pyunits,
                           value,
                           Var)
from pyomo.util.check_units import assert_units_consistent
from pyomo.common.config import ConfigBlock, ConfigValue

from idaes.core import FlowsheetBlock
from idaes.core.util import get_solver
from idaes.generic_models.costing import UnitModelCostingBlock
from idaes.core.util.model_statistics import \
    degrees_of_freedom, number_unfixed_variables
from idaes.core.util.misc import add_object_reference

from watertap.core.zero_order_costing import \
    ZeroOrderCosting,  _load_case_study_definition, _get_tech_parameters
from watertap.core.zero_order_properties import WaterParameterBlock
from watertap.core.wt_database import Database
from watertap.unit_models.zero_order import \
    ChemicalAdditionZO, NanofiltrationZO

solver = get_solver()


# TODO: Test general costing methods

class TestGeneralMethods:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()

        m.frame = ZeroOrderCosting()

        # Dummy a unit model to use with _get_tech_parameters
        m.dummy_unit = Block(concrete=True)
        m.dummy_unit.config = ConfigBlock()
        m.dummy_unit.config.declare("flowsheet_costing_block", ConfigValue())
        m.dummy_unit.config.flowsheet_costing_block = m.frame
        add_object_reference(m.dummy_unit, "unit_model", m.dummy_unit)
        m.dummy_unit._tech_type = "test_tech"

        return m

    @pytest.mark.unit
    def test_load_case_study_definition(self, model):
        assert model.frame.config.case_study_definition is None

        def_dict = _load_case_study_definition(model.frame)

        assert isinstance(def_dict, dict)

        model.frame.config.case_study_definition = "foo"
        with pytest.raises(OSError,
                           match="Could not find specified case study "
                           "definition file. Please check the path provided."):
            _load_case_study_definition(model.frame)

    @pytest.mark.unit
    def test_build_global_params(self, model):
        assert model.frame.base_currency == pyunits.MUSD_2018
        assert model.frame.base_period == pyunits.year

        assert len(model.frame.defined_flows) == 20
        for f in model.frame.defined_flows:
            assert f in ["electricity",
                         "acetic_acid",
                         "activated_carbon",
                         "alum",
                         "ammonia",
                         "anthracite",
                         "anti-scalant",
                         "cationic_polymer",
                         "caustic_soda",
                         "chlorine",
                         "ferric_chloride",
                         "hydrochloric_acid",
                         "hydrogen_peroxide",
                         "ion_exchange_resin",
                         "lime",
                         "phosphoric_acid",
                         "sand",
                         "sodium_bisulfite",
                         "sodium_chloride",
                         "sulfuric_acid"]

        assert number_unfixed_variables(model.frame) == 0

        assert isinstance(model.frame.plant_lifetime, Var)
        assert value(model.frame.plant_lifetime) == 30
        assert isinstance(model.frame.utilization_factor, Var)
        assert value(model.frame.utilization_factor) == 1

        assert isinstance(model.frame.land_cost_percent_FCI, Var)
        assert value(model.frame.land_cost_percent_FCI) == 0.0015
        assert isinstance(model.frame.working_capital_percent_FCI, Var)
        assert value(model.frame.working_capital_percent_FCI) == 0.05
        assert isinstance(model.frame.salaries_percent_FCI, Var)
        assert value(model.frame.salaries_percent_FCI) == 0.001
        assert isinstance(model.frame.benefit_percent_of_salary, Var)
        assert value(model.frame.benefit_percent_of_salary) == 0.9
        assert isinstance(model.frame.maintenance_costs_percent_FCI, Var)
        assert value(model.frame.maintenance_costs_percent_FCI) == 0.008
        assert isinstance(model.frame.laboratory_fees_percent_FCI, Var)
        assert value(model.frame.laboratory_fees_percent_FCI) == 0.003
        assert isinstance(model.frame.insurance_and_taxes_percent_FCI, Var)
        assert value(model.frame.insurance_and_taxes_percent_FCI) == 0.002

        assert isinstance(model.frame.wacc, Var)
        assert value(model.frame.wacc) == 0.05
        assert isinstance(model.frame.capital_recovery_factor, Expression)
        assert value(model.frame.capital_recovery_factor) == pytest.approx(
            0.0650514, rel=1e-5)

        assert isinstance(model.frame.TPEC, Var)
        assert value(model.frame.TPEC) == 3.4
        assert isinstance(model.frame.TIC, Var)
        assert value(model.frame.TIC) == 1.65

    @pytest.mark.unit
    def test_get_tech_parameters_first_call(self, model):
        assert not hasattr(model.frame, "test_tech")

        parameters = {
            "capital_cost": {
                "capital_a_parameter": {
                    "value": 1,
                    "units": "m"},
                "capital_b_parameter": {
                    "value": 7,
                    "units": "K"},
                "reference_state": {
                    "value": 22,
                    "units": "mol"}}}

        A, B, R = _get_tech_parameters(
            model.dummy_unit, parameters, None, ["capital_a_parameter",
                                                 "capital_b_parameter",
                                                 "reference_state"])

        assert isinstance(model.frame.test_tech, Block)
        assert isinstance(model.frame.test_tech.subtype_set, Set)
        assert len(model.frame.test_tech.subtype_set) == 1
        assert None in model.frame.test_tech.subtype_set

        assert isinstance(model.frame.test_tech.capital_a_parameter, Var)
        assert model.frame.test_tech.capital_a_parameter.index_set() is \
            model.frame.test_tech.subtype_set
        assert value(model.frame.test_tech.capital_a_parameter[None]) == 1
        assert model.frame.test_tech.capital_a_parameter[None].fixed

        assert isinstance(model.frame.test_tech.capital_b_parameter, Var)
        assert model.frame.test_tech.capital_b_parameter.index_set() is \
            model.frame.test_tech.subtype_set
        assert value(model.frame.test_tech.capital_b_parameter[None]) == 7
        assert model.frame.test_tech.capital_b_parameter[None].fixed

        assert isinstance(model.frame.test_tech.reference_state, Var)
        assert model.frame.test_tech.reference_state.index_set() is \
            model.frame.test_tech.subtype_set
        assert value(model.frame.test_tech.reference_state[None]) == 22
        assert model.frame.test_tech.reference_state[None].fixed

        assert A is model.frame.test_tech.capital_a_parameter[None]
        assert B is model.frame.test_tech.capital_b_parameter[None]
        assert R is model.frame.test_tech.reference_state[None]

    @pytest.mark.unit
    def test_get_tech_parameters_second_call(self, model):
        # Use differnt parameter values- thes eshould be ignored
        parameters = {
            "capital_cost": {
                "capital_a_parameter": {
                    "value": 100,
                    "units": "m"},
                "capital_b_parameter": {
                    "value": 700,
                    "units": "K"},
                "reference_state": {
                    "value": 2200,
                    "units": "mol"}}}

        A, B, R = _get_tech_parameters(
            model.dummy_unit, parameters, None, ["capital_a_parameter",
                                                 "capital_b_parameter",
                                                 "reference_state"])

        assert isinstance(model.frame.test_tech, Block)
        assert isinstance(model.frame.test_tech.subtype_set, Set)
        assert len(model.frame.test_tech.subtype_set) == 1
        assert None in model.frame.test_tech.subtype_set

        assert isinstance(model.frame.test_tech.capital_a_parameter, Var)
        assert model.frame.test_tech.capital_a_parameter.index_set() is \
            model.frame.test_tech.subtype_set
        assert value(model.frame.test_tech.capital_a_parameter[None]) == 1
        assert model.frame.test_tech.capital_a_parameter[None].fixed

        assert isinstance(model.frame.test_tech.capital_b_parameter, Var)
        assert model.frame.test_tech.capital_b_parameter.index_set() is \
            model.frame.test_tech.subtype_set
        assert value(model.frame.test_tech.capital_b_parameter[None]) == 7
        assert model.frame.test_tech.capital_b_parameter[None].fixed

        assert isinstance(model.frame.test_tech.reference_state, Var)
        assert model.frame.test_tech.reference_state.index_set() is \
            model.frame.test_tech.subtype_set
        assert value(model.frame.test_tech.reference_state[None]) == 22
        assert model.frame.test_tech.reference_state[None].fixed

        assert A is model.frame.test_tech.capital_a_parameter[None]
        assert B is model.frame.test_tech.capital_b_parameter[None]
        assert R is model.frame.test_tech.reference_state[None]

    @pytest.mark.unit
    def test_get_tech_parameters_second_subtype(self, model):
        # Different parameters should be used this time
        parameters = {
            "capital_cost": {
                "capital_a_parameter": {
                    "value": 100,
                    "units": "m"},
                "capital_b_parameter": {
                    "value": 700,
                    "units": "K"},
                "reference_state": {
                    "value": 2200,
                    "units": "mol"}}}

        # Provide subtype foo
        A, B, R = _get_tech_parameters(
            model.dummy_unit, parameters, "foo", ["capital_a_parameter",
                                                  "capital_b_parameter",
                                                  "reference_state"])

        assert isinstance(model.frame.test_tech, Block)
        assert isinstance(model.frame.test_tech.subtype_set, Set)
        assert len(model.frame.test_tech.subtype_set) == 2
        assert "foo" in model.frame.test_tech.subtype_set

        assert isinstance(model.frame.test_tech.capital_a_parameter, Var)
        assert model.frame.test_tech.capital_a_parameter.index_set() is \
            model.frame.test_tech.subtype_set
        assert value(model.frame.test_tech.capital_a_parameter["foo"]) == 100
        assert model.frame.test_tech.capital_a_parameter["foo"].fixed

        assert isinstance(model.frame.test_tech.capital_b_parameter, Var)
        assert model.frame.test_tech.capital_b_parameter.index_set() is \
            model.frame.test_tech.subtype_set
        assert value(model.frame.test_tech.capital_b_parameter["foo"]) == 700
        assert model.frame.test_tech.capital_b_parameter["foo"].fixed

        assert isinstance(model.frame.test_tech.reference_state, Var)
        assert model.frame.test_tech.reference_state.index_set() is \
            model.frame.test_tech.subtype_set
        assert value(model.frame.test_tech.reference_state["foo"]) == 2200
        assert model.frame.test_tech.reference_state["foo"].fixed

        assert A is model.frame.test_tech.capital_a_parameter["foo"]
        assert B is model.frame.test_tech.capital_b_parameter["foo"]
        assert R is model.frame.test_tech.reference_state["foo"]


class TestWorkflow:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(default={"dynamic": False})

        m.fs.params = WaterParameterBlock(
            default={"solute_list": ["sulfur", "toc", "tss"]})

        m.fs.costing = ZeroOrderCosting()

        return m

    @pytest.mark.component
    def test_nf_costing(self, model):
        model.fs.unit1 = NanofiltrationZO(default={
            "property_package": model.fs.params,
            "database": model.db})

        model.fs.unit1.inlet.flow_mass_comp[0, "H2O"].fix(10000)
        model.fs.unit1.inlet.flow_mass_comp[0, "sulfur"].fix(1)
        model.fs.unit1.inlet.flow_mass_comp[0, "toc"].fix(2)
        model.fs.unit1.inlet.flow_mass_comp[0, "tss"].fix(3)
        model.fs.unit1.load_parameters_from_database()
        assert degrees_of_freedom(model.fs.unit1) == 0

        model.fs.unit1.costing = UnitModelCostingBlock(default={
            "flowsheet_costing_block": model.fs.costing})

        assert isinstance(model.fs.costing.nanofiltration, Block)
        assert isinstance(model.fs.costing.nanofiltration.capital_a_parameter,
                          Var)
        assert isinstance(model.fs.costing.nanofiltration.capital_b_parameter,
                          Var)
        assert isinstance(model.fs.costing.nanofiltration.reference_state, Var)

        assert isinstance(model.fs.unit1.costing.capital_cost, Var)
        assert isinstance(model.fs.unit1.costing.capital_cost_constraint,
                          Constraint)

        assert_units_consistent(model.fs)
        assert degrees_of_freedom(model.fs.unit1) == 0

        assert model.fs.unit1.electricity[0] in \
            model.fs.costing._registered_flows["electricity"]

    @pytest.mark.component
    def test_chem_addition_costing(self, model):
        model.fs.unit2 = ChemicalAdditionZO(default={
            "property_package": model.fs.params,
            "process_subtype": "alum",
            "database": model.db})

        model.fs.unit2.inlet.flow_mass_comp[0, "H2O"].fix(1000)
        model.fs.unit2.inlet.flow_mass_comp[0, "sulfur"].fix(1)
        model.fs.unit2.inlet.flow_mass_comp[0, "toc"].fix(2)
        model.fs.unit2.inlet.flow_mass_comp[0, "tss"].fix(3)
        model.fs.unit2.load_parameters_from_database()
        assert degrees_of_freedom(model.fs.unit2) == 0

        model.fs.unit2.costing = UnitModelCostingBlock(default={
            "flowsheet_costing_block": model.fs.costing})

        assert isinstance(
            model.fs.costing.chemical_addition, Block)
        assert isinstance(
            model.fs.costing.chemical_addition.capital_a_parameter, Var)
        assert isinstance(
            model.fs.costing.chemical_addition.capital_b_parameter, Var)

        assert isinstance(model.fs.unit2.costing.capital_cost, Var)
        assert isinstance(model.fs.unit2.costing.capital_cost_constraint,
                          Constraint)

        assert_units_consistent(model.fs)
        assert degrees_of_freedom(model.fs.unit2) == 0

        assert model.fs.unit2.electricity[0] is \
            model.fs.costing._registered_flows["electricity"][1]
        assert pytest.approx(1.006e3*10/0.5, rel=1e-8) == value(pyunits.convert(
            model.fs.costing._registered_flows["alum"][0],
            to_units=pyunits.mg/pyunits.s))

    @pytest.mark.component
    def test_process_costing(self, model):
        model.fs.costing.cost_process()

        assert isinstance(model.fs.costing.aggregate_capital_cost, Var)
        assert isinstance(model.fs.costing.aggregate_fixed_operating_cost, Var)
        assert isinstance(model.fs.costing.aggregate_variable_operating_cost,
                          Var)
        assert isinstance(model.fs.costing.aggregate_flow_electricity,
                          Var)
        assert isinstance(model.fs.costing.aggregate_flow_costs,
                          Var)

        assert isinstance(model.fs.costing.land_cost, Var)
        assert isinstance(model.fs.costing.working_capital, Var)
        assert isinstance(model.fs.costing.total_capital_cost, Var)

        assert isinstance(model.fs.costing.land_cost_constraint, Constraint)
        assert isinstance(model.fs.costing.working_capital_constraint,
                          Constraint)
        assert isinstance(model.fs.costing.total_capital_cost_constraint,
                          Constraint)

        assert_units_consistent(model.fs)
        assert degrees_of_freedom(model.fs) == 0

    @pytest.mark.component
    def test_add_LCOW(self, model):
        model.fs.costing.add_LCOW(model.fs.unit1.properties_in[0].flow_vol)

        assert isinstance(model.fs.costing.LCOW, Expression)

        assert_units_consistent(model.fs)
        assert degrees_of_freedom(model.fs) == 0

    @pytest.mark.component
    def test_add_electricity_intensity(self, model):
        model.fs.costing.add_electricity_intensity(
            model.fs.unit1.properties_in[0].flow_vol)

        assert isinstance(model.fs.costing.electricity_intensity, Expression)

        assert_units_consistent(model.fs)
        assert degrees_of_freedom(model.fs) == 0

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, model):
        model.fs.unit1.initialize()
        model.fs.unit2.initialize()

        model.fs.costing.initialize()

        assert pytest.approx(0, abs=1e-5) == value(
            model.fs.costing.land_cost_constraint)
        assert pytest.approx(0, abs=1e-5) == value(
            model.fs.costing.working_capital_constraint)
        assert pytest.approx(0, abs=1e-5) == value(
            model.fs.costing.total_capital_cost_constraint)

        assert pytest.approx(0, abs=1e-5) == value(
            model.fs.costing.salary_cost_constraint)
        assert pytest.approx(0, abs=1e-5) == value(
            model.fs.costing.benefits_cost_constraint)
        assert pytest.approx(0, abs=1e-5) == value(
            model.fs.costing.maintenance_cost_constraint)
        assert pytest.approx(0, abs=1e-5) == value(
            model.fs.costing.laboratory_cost_constraint)
        assert pytest.approx(0, abs=1e-5) == value(
            model.fs.costing.insurance_and_taxes_cost_constraint)

        assert pytest.approx(0, abs=1e-5) == value(
            model.fs.costing.total_fixed_operating_cost_constraint)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model, tee=True)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        # Note all dollar values are in millions of dollars
        assert pytest.approx(630.031, rel=1e-5) == value(
            model.fs.costing.total_capital_cost)

        assert pytest.approx(8333.42, rel=1e-5) == value(
            model.fs.costing.aggregate_flow_electricity)
        assert pytest.approx(20.12, rel=1e-5) == value(
            model.fs.costing.aggregate_flow_alum)

        # Note units (M$)
        assert pytest.approx(1.73136e-7, rel=1e-5) == value(
            model.fs.costing.LCOW)

        assert pytest.approx(0.231345, rel=1e-5) == value(
            model.fs.costing.electricity_intensity)



###############################################################################
# ProteusLib Copyright (c) 2021, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National
# Laboratory, National Renewable Energy Laboratory, and National Energy
# Technology Laboratory (subject to receipt of any required approvals from
# the U.S. Dept. of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/nawi-hub/proteuslib/"
#
###############################################################################
import re

import pytest
from pyomo.environ import (ConcreteModel,
                           assert_optimal_termination,
                           Set,
                           Param,
                           units as pyunits,
                           Suffix)
from idaes.core import (FlowsheetBlock,
                        MaterialFlowBasis,
                        MaterialBalanceType,
                        )
from idaes.core.util.scaling import calculate_scaling_factors, get_scaling_factor

from pyomo.util.check_units import assert_units_consistent
from watertap.property_models.ion_DSPMDE_prop_pack import DSPMDEParameterBlock, DSPMDEStateBlock, \
    ActivityCoefficientModel, DensityCalculation
from watertap.core.util.initialization import check_dof
from idaes.core.util.model_statistics import *
from idaes.core.util.scaling import unscaled_variables_generator, unscaled_constraints_generator, badly_scaled_var_generator
from watertap.property_models.tests.property_test_harness import PropertyAttributeError
from idaes.core.util import get_solver

solver = get_solver()
# -----------------------------------------------------------------------------

@pytest.fixture(scope="module")
def model():
    m = ConcreteModel()

    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.properties = DSPMDEParameterBlock(default={
        "solute_list": ["A", "B", "C", "D"],
        "diffusivity_data": {("Liq", "A"): 1e-9,
                             ("Liq", "B"): 1e-10,
                             ("Liq", "C"): 1e-7,
                             ("Liq", "D"): 1e-11},
        "mw_data": {"H2O": 18e-3,
                    "A": 10e-3,
                    "B": 25e-3,
                    "C": 100e-3,
                    "D": 25e-3},
        "stokes_radius_data": {"A": 1e-9,
                               "B": 1e-9,
                               "C": 1e-9,
                               "D": 1e-10},
        "charge": {"A": 1,
                   "B": -2,
                   "C": 2,
                   "D": -1}
    })

    return m

@pytest.mark.unit
def test_parameter_block(model):
    assert isinstance(model.fs.properties.component_list, Set)
    for j in model.fs.properties.component_list:
        assert j in ["H2O", "A", "B", "C", "D"]
    assert isinstance(model.fs.properties.solvent_set, Set)
    for j in model.fs.properties.solvent_set:
        assert j in ["H2O"]
    assert isinstance(model.fs.properties.solute_set, Set)
    for j in model.fs.properties.solute_set:
        assert j in ["A", "B", "C", "D"]

    assert isinstance(model.fs.properties.phase_list, Set)
    for j in model.fs.properties.phase_list:
        assert j in ["Liq"]

    assert model.fs.properties._state_block_class is DSPMDEStateBlock

    assert isinstance(model.fs.properties.mw_comp, Param)
    assert model.fs.properties.mw_comp['A'].value == 10e-3
    assert model.fs.properties.mw_comp['B'].value == 25e-3
    assert model.fs.properties.mw_comp['C'].value == 100e-3
    assert model.fs.properties.mw_comp['H2O'].value == 18e-3

    assert isinstance(model.fs.properties.diffus_phase_comp, Param)
    assert model.fs.properties.diffus_phase_comp['Liq', 'A'].value == 1e-9
    assert model.fs.properties.diffus_phase_comp['Liq', 'B'].value == 1e-10
    assert model.fs.properties.diffus_phase_comp['Liq', 'C'].value == 1e-7

    assert isinstance(model.fs.properties.radius_stokes_comp, Param)
    assert model.fs.properties.radius_stokes_comp['A'].value == 1e-9
    assert model.fs.properties.radius_stokes_comp['B'].value == 1e-9
    assert model.fs.properties.radius_stokes_comp['C'].value == 1e-9

    assert model.fs.properties.config.activity_coefficient_model == ActivityCoefficientModel.ideal


@pytest.mark.component
def test_property_ions(model):
    m = model
    m.fs.stream = m.fs.properties.build_state_block([0], default={'defined_state': True})

    m.fs.stream[0].flow_mol_phase_comp['Liq', 'A'].fix(0.000407)
    m.fs.stream[0].flow_mol_phase_comp['Liq', 'B'].fix(0.010479)
    m.fs.stream[0].flow_mol_phase_comp['Liq', 'C'].fix(0.010479)
    m.fs.stream[0].flow_mol_phase_comp['Liq', 'D'].fix(0.000407)
    m.fs.stream[0].flow_mol_phase_comp['Liq', 'H2O'].fix(0.99046)
    m.fs.stream[0].temperature.fix(298.15)
    m.fs.stream[0].pressure.fix(101325)

    m.fs.stream[0].assert_electroneutrality(defined_state=True)

    m.fs.stream[0].mole_frac_phase_comp

    m.fs.stream[0].flow_mass_phase_comp

    m.fs.stream[0].molality_comp
    m.fs.stream[0].pressure_osm
    m.fs.stream[0].dens_mass_phase
    m.fs.stream[0].conc_mol_phase_comp
    m.fs.stream[0].act_coeff_phase_comp

    calculate_scaling_factors(m.fs)

    assert_units_consistent(m)

    check_dof(m, fail_flag=True)

    m.fs.stream.initialize()

    results = solver.solve(m)
    assert_optimal_termination(results)

    assert value(m.fs.stream[0].conc_mass_phase_comp['Liq','A']) == pytest.approx(2.1288e-1,  rel=1e-3)

    assert value(m.fs.stream[0].conc_mol_phase_comp['Liq','A']) == pytest.approx(21.288,  rel=1e-3)
    assert value(m.fs.stream[0].molality_comp['A']) == pytest.approx(2.2829e-2,  rel=1e-3)

    assert value(m.fs.stream[0].pressure_osm) == pytest.approx(60.546e5,  rel=1e-3)

    assert value(m.fs.stream[0].dens_mass_phase['Liq']) == pytest.approx(1001.76,  rel=1e-3)

    assert value(m.fs.stream[0].act_coeff_phase_comp['Liq', 'A']) == 1

@pytest.fixture(scope="module")
def model2():
    m = ConcreteModel()

    m.fs = FlowsheetBlock(default={"dynamic": False})
    m.fs.properties = DSPMDEParameterBlock(default={
        "solute_list": ["A", "B", "C", "D"]})

    return m

@pytest.mark.component
def test_property_ions(model2):
    m = model2

    stream = m.fs.stream = m.fs.properties.build_state_block([0], default={'defined_state': True})

    stream[0].flow_mol_phase_comp['Liq', 'A'].fix(0.000407)
    stream[0].flow_mol_phase_comp['Liq', 'B'].fix(0.010479)
    stream[0].flow_mol_phase_comp['Liq', 'C'].fix(0.010479)
    stream[0].flow_mol_phase_comp['Liq', 'D'].fix(0.000407)
    stream[0].flow_mol_phase_comp['Liq', 'H2O'].fix(0.99046)
    stream[0].temperature.fix(298.15)
    stream[0].pressure.fix(101325)

    stream[0].diffus_phase_comp['Liq', 'A'] = 1e-9
    stream[0].diffus_phase_comp['Liq', 'B'] = 1e-10
    stream[0].diffus_phase_comp['Liq', 'C'] = 1e-7
    stream[0].diffus_phase_comp['Liq', 'D'] = 1e-11

    stream[0].mw_comp['H2O'] = 18e-3
    stream[0].mw_comp['A'] = 10e-3
    stream[0].mw_comp['B'] = 25e-3
    stream[0].mw_comp['C'] = 100e-3
    stream[0].mw_comp['D'] = 25e-3

    stream[0].radius_stokes_comp['A'] = 1e-9
    stream[0].radius_stokes_comp['B'] = 1e-9
    stream[0].radius_stokes_comp['C'] = 1e-9
    stream[0].radius_stokes_comp['D'] = 1e-10

    stream[0].charge_comp['A'] = 1
    stream[0].charge_comp['B'] = -2
    stream[0].charge_comp['C'] = 2
    stream[0].charge_comp['D'] = -1

    stream[0].assert_electroneutrality(defined_state=True, tol=1e-8)

    stream[0].mole_frac_phase_comp

    stream[0].flow_mass_phase_comp

    stream[0].molality_comp
    stream[0].pressure_osm
    stream[0].dens_mass_phase
    stream[0].conc_mol_phase_comp
    stream[0].flow_vol
    calculate_scaling_factors(m.fs)

    assert_units_consistent(m)

    check_dof(m, fail_flag=True)

    stream.initialize()

    results = solver.solve(m)
    assert_optimal_termination(results)
    assert value(stream[0].flow_vol_phase['Liq']) == pytest.approx(1.91524e-5,  rel=1e-3)


@pytest.fixture(scope="module")
def model3():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={"dynamic": False})

    m.fs.properties = DSPMDEParameterBlock(default={
        "solute_list": ["Ca_2+", "SO4_2-", "Na_+", "Cl_-", "Mg_2+"]
        })

    m.fs.stream = m.fs.properties.build_state_block([0], default={'defined_state': True})

    return m

@pytest.mark.unit
def test_build(model3):
    m = model3

    # test scaling factor
    assert hasattr(m.fs.stream[0], 'scaling_factor')
    assert isinstance(m.fs.stream[0].scaling_factor, Suffix)

    # test state variables
    state_vars_list = ['flow_mol_phase_comp', 'temperature', 'pressure']
    state_vars_dict = m.fs.stream[0].define_state_vars()
    assert len(state_vars_dict) == len(state_vars_list)
    for sv in state_vars_list:
        assert sv in state_vars_dict
        assert hasattr(m.fs.stream[0], sv)
        var = getattr(m.fs.stream[0], sv)
        assert isinstance(var, Var)

    metadata = m.fs.properties.get_metadata().properties

    # check that properties are not built if not demanded
    for v_name in metadata:
        if metadata[v_name]['method'] is not None:
            if m.fs.stream[0].is_property_constructed(v_name):
                raise PropertyAttributeError(
                    "Property {v_name} is an on-demand property, but was found "
                    "on the stateblock without being demanded".format(v_name=v_name))

    # check that properties are built if demanded
    for v_name in metadata:
        if metadata[v_name]['method'] is not None:
            if not hasattr(m.fs.stream[0], v_name):
                raise PropertyAttributeError(
                    "Property {v_name} is an on-demand property, but was not built "
                    "when demanded".format(v_name=v_name))

    assert m.fs.stream[0].is_property_constructed('act_coeff_phase_comp')

    var_list = ['mass_frac_phase_comp', 'dens_mass_phase', 'flow_vol_phase',
                'conc_mass_phase_comp', 'flow_mass_phase_comp', 'mole_frac_phase_comp',
                'molality_comp', 'pressure_osm', 'act_coeff_phase_comp']

    # test on demand constraints
    for v in var_list:
        assert hasattr(m.fs.stream[0], '_' + v)  # check method
        assert hasattr(m.fs.stream[0], 'eq_' + v)  # check constraint
        c = getattr(m.fs.stream[0], 'eq_' + v)
        assert isinstance(c, Constraint)

    assert number_variables(m) == 64
    assert number_total_constraints(m) == 46
    assert number_unused_variables(m) == 6

@pytest.mark.unit
def test_general_methods(model3):
    m = model3

    assert hasattr(m.fs.stream[0], 'get_material_flow_terms')

    assert hasattr(m.fs.stream[0], 'default_material_balance_type')
    assert (m.fs.stream[0].default_material_balance_type()
            is MaterialBalanceType.componentTotal)

    assert hasattr(m.fs.stream[0], 'get_material_flow_basis')
    assert (m.fs.stream[0].get_material_flow_basis()
            is MaterialFlowBasis.molar)

@pytest.mark.unit
def test_default_scaling(model3):
    m = model3

    assert hasattr(m.fs.properties, 'default_scaling_factor')
    default_scaling_var_dict = {('temperature', None): 1e-2,
                                ('pressure', None): 1e-6,
                                ('dens_mass_phase', 'Liq'): 1e-3,
                                ('visc_d_phase', 'Liq'): 1e3,
                                ('diffus_phase_comp', 'Liq'): 1e10,}

    assert len(default_scaling_var_dict) == len(m.fs.properties.default_scaling_factor)
    for t, sf in default_scaling_var_dict.items():
        assert t in m.fs.properties.default_scaling_factor.keys()
        assert m.fs.properties.default_scaling_factor[t] == sf

@pytest.mark.unit
def test_scaling(model3):
    m = model3
    metadata = m.fs.properties.get_metadata().properties

    for v_name in metadata:
        getattr(m.fs.stream[0], v_name)

    calculate_scaling_factors(m)

    # check that all variables have scaling factors
    unscaled_var_list = list(unscaled_variables_generator(m))
    [print(i) for i in unscaled_var_list]
    assert len(unscaled_var_list) == 0

    # check if any variables are badly scaled
    badly_scaled_var_list = list(badly_scaled_var_generator(m))
    assert len(badly_scaled_var_list) == 0

    # check that all constraints have been scaled
    unscaled_constraint_list = list(unscaled_constraints_generator(m))
    [print(i) for i in unscaled_constraint_list]

    assert len(unscaled_constraint_list) == 0

    # m.fs.stream[0].scaling_factor.display()
    for j in m.fs.properties.config.solute_list:
        assert get_scaling_factor(m.fs.stream[0].mw_comp[j]) is not None
        assert get_scaling_factor(m.fs.stream[0].diffus_phase_comp['Liq', j]) is not None
        assert get_scaling_factor(m.fs.stream[0].act_coeff_phase_comp['Liq', j]) is not None
    assert get_scaling_factor(m.fs.stream[0].dens_mass_phase['Liq']) is not None
    assert get_scaling_factor(m.fs.stream[0].visc_d_phase['Liq']) is not None

@pytest.mark.component
def test_seawater_data():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={'dynamic':False})
    m.fs.properties = DSPMDEParameterBlock(default={
        "solute_list": ["Ca_2+", "SO4_2-", "Na_+", "Cl_-", "Mg_2+"],
        "diffusivity_data": {("Liq", "Ca_2+"): 0.792e-9,
                             ("Liq", "SO4_2-"): 1.06e-9,
                             ("Liq", "Na_+"): 1.33e-9,
                             ("Liq", "Cl_-"): 2.03e-9,
                             ("Liq", "Mg_2+"): 0.706e-9},
        "mw_data": {"H2O": 18e-3,
                    "Na_+": 23e-3,
                    "Ca_2+": 40e-3,
                    "Mg_2+": 24e-3,
                    "Cl_-": 35e-3,
                    "SO4_2-": 96e-3},
        "stokes_radius_data": {"Na_+": 0.184e-9,
                               "Ca_2+": 0.309e-9,
                               "Mg_2+": 0.347e-9,
                               "Cl_-": 0.121e-9,
                               "SO4_2-": 0.230e-9},
        "charge": {"Na_+": 1,
                   "Ca_2+": 2,
                   "Mg_2+": 2,
                   "Cl_-": -1,
                   "SO4_2-": -2},
        "density_calculation": DensityCalculation.seawater,
        "activity_coefficient_model": ActivityCoefficientModel.davies
    })

    m.fs.stream = stream = m.fs.properties.build_state_block([0], default={'defined_state': True})

    mass_flow_in = 1 * pyunits.kg / pyunits.s
    feed_mass_frac = {'Na_+': 11122e-6,
                      'Ca_2+': 382e-6,
                      'Mg_2+': 1394e-6,
                      'SO4_2-': 2136e-6,
                      'Cl_-': 20300e-6}
    for ion, x in feed_mass_frac.items():
        mol_comp_flow = x * pyunits.kg / pyunits.kg * mass_flow_in / stream[0].mw_comp[ion]

        stream[0].flow_mol_phase_comp['Liq', ion].fix(mol_comp_flow)

    H2O_mass_frac = 1 - sum(x for x in feed_mass_frac.values())
    H2O_mol_comp_flow = H2O_mass_frac * pyunits.kg / pyunits.kg * mass_flow_in / stream[0].mw_comp['H2O']

    stream[0].flow_mol_phase_comp['Liq', 'H2O'].fix(H2O_mol_comp_flow)
    stream[0].temperature.fix(298.15)
    stream[0].pressure.fix(101325)

    stream[0].assert_electroneutrality(defined_state=True, tol=1e-2, adjust_by_ion='Cl_-')

    metadata = m.fs.properties.get_metadata().properties
    for v_name in metadata:
        getattr(stream[0], v_name)
    assert stream[0].is_property_constructed('conc_mol_phase_comp')

    assert_units_consistent(m)

    check_dof(m, fail_flag=True)

    m.fs.properties.set_default_scaling('flow_mol_phase_comp', 1, index=('Liq', 'H2O'))
    m.fs.properties.set_default_scaling('flow_mol_phase_comp', 1, index=('Liq', 'Na_+'))
    m.fs.properties.set_default_scaling('flow_mol_phase_comp', 1, index=('Liq', 'Cl_-'))
    m.fs.properties.set_default_scaling('flow_mol_phase_comp', 1e4, index=('Liq', 'Ca_2+'))
    m.fs.properties.set_default_scaling('flow_mol_phase_comp', 1e4, index=('Liq', 'SO4_2-'))
    m.fs.properties.set_default_scaling('flow_mol_phase_comp', 1, index=('Liq', 'Mg_2+'))
    calculate_scaling_factors(m)

    # check if any variables are badly scaled
    badly_scaled_var_list = list(badly_scaled_var_generator(m))
    assert len(badly_scaled_var_list) == 0

    stream.initialize()

    # check if any variables are badly scaled
    badly_scaled_var_list = list(badly_scaled_var_generator(m))
    assert len(badly_scaled_var_list) == 0

    results = solver.solve(m)
    assert_optimal_termination(results)

    assert value(stream[0].flow_vol_phase['Liq']) == pytest.approx(9.767e-4,  rel=1e-3)
    assert value(stream[0].flow_mol_phase_comp['Liq', 'H2O']) == pytest.approx(53.59256,  rel=1e-3)
    assert value(stream[0].flow_mol_phase_comp['Liq', 'Na_+']) == pytest.approx(0.4836,  rel=1e-3)
    assert value(stream[0].flow_mol_phase_comp['Liq', 'Ca_2+']) == pytest.approx(0.00955,  rel=1e-3)
    assert value(stream[0].flow_mol_phase_comp['Liq', 'Mg_2+']) == pytest.approx(0.05808,  rel=1e-3)
    assert value(stream[0].flow_mol_phase_comp['Liq', 'Cl_-']) == pytest.approx(0.57443,  rel=1e-3)
    assert value(stream[0].flow_mol_phase_comp['Liq', 'SO4_2-']) == pytest.approx(0.02225,  rel=1e-3)

    assert value(stream[0].dens_mass_phase['Liq']) == pytest.approx(1023.816, rel=1e-3)
    assert value(stream[0].pressure_osm) == pytest.approx(29.132e5, rel=1e-3)
    assert value(stream[0].flow_vol) == pytest.approx(9.767e-4, rel=1e-3)

    assert value(sum(stream[0].conc_mass_phase_comp['Liq', j]
                     for j in m.fs.properties.solute_set)) == pytest.approx(35.9744, rel=1e-3)
    assert value(sum(stream[0].mass_frac_phase_comp['Liq', j]
                     for j in m.fs.properties.solute_set)) == pytest.approx(0.035142, rel=1e-3)
    assert value(sum(stream[0].mass_frac_phase_comp['Liq', j]
                     for j in m.fs.properties.component_list)) == pytest.approx(1, rel=1e-3)
    assert value(sum(stream[0].mole_frac_phase_comp['Liq', j]
                     for j in m.fs.properties.component_list)) == pytest.approx(1, rel=1e-3)

    assert value(stream[0].conc_mol_phase_comp['Liq', 'Na_+']) == pytest.approx(495.082, rel=1e-3)
    assert value(stream[0].conc_mol_phase_comp['Liq', 'Cl_-']) == pytest.approx(588.0431, rel=1e-3)
    assert value(stream[0].conc_mol_phase_comp['Liq', 'Ca_2+']) == pytest.approx(9.777, rel=1e-3)
    assert value(stream[0].conc_mol_phase_comp['Liq', 'SO4_2-']) == pytest.approx(22.780, rel=1e-3)
    assert value(stream[0].conc_mol_phase_comp['Liq', 'Mg_2+']) == pytest.approx(59.467, rel=1e-3)

    assert value(stream[0].conc_mass_phase_comp['Liq', 'Na_+']) == pytest.approx(11.387, rel=1e-3)
    assert value(stream[0].conc_mass_phase_comp['Liq', 'Cl_-']) == pytest.approx(20.5815, rel=1e-3)
    assert value(stream[0].conc_mass_phase_comp['Liq', 'Ca_2+']) == pytest.approx(0.391, rel=1e-3)
    assert value(stream[0].conc_mass_phase_comp['Liq', 'SO4_2-']) == pytest.approx(2.187, rel=1e-3)
    assert value(stream[0].conc_mass_phase_comp['Liq', 'Mg_2+']) == pytest.approx(1.427, rel=1e-3)

    assert value(stream[0].mole_frac_phase_comp['Liq', 'Na_+']) == pytest.approx(8.833e-3, rel=1e-3)
    assert value(stream[0].mole_frac_phase_comp['Liq', 'Cl_-']) == pytest.approx(1.049e-2, rel=1e-3)
    assert value(stream[0].mole_frac_phase_comp['Liq', 'Ca_2+']) == pytest.approx(1.744e-4, rel=1e-3)
    assert value(stream[0].mole_frac_phase_comp['Liq', 'SO4_2-']) == pytest.approx(4.064e-4, rel=1e-3)
    assert value(stream[0].mole_frac_phase_comp['Liq', 'Mg_2+']) == pytest.approx(1.061e-3, rel=1e-3)

    assert value(stream[0].mass_frac_phase_comp['Liq', 'Na_+']) == pytest.approx(1.112e-2, rel=1e-3)
    assert value(stream[0].mass_frac_phase_comp['Liq', 'Cl_-']) == pytest.approx(2.01e-2, rel=1e-3)
    assert value(stream[0].mass_frac_phase_comp['Liq', 'Ca_2+']) == pytest.approx(3.82e-4, rel=1e-3)
    assert value(stream[0].mass_frac_phase_comp['Liq', 'SO4_2-']) == pytest.approx(2.136e-3, rel=1e-3)
    assert value(stream[0].mass_frac_phase_comp['Liq', 'Mg_2+']) == pytest.approx(1.394e-3, rel=1e-3)

    assert value(stream[0].debye_huckel_constant) == pytest.approx(0.01554, rel=1e-3)
    assert value(stream[0].ionic_strength) == pytest.approx(0.73467, rel=1e-3)

@pytest.mark.requires_idaes_solver
@pytest.mark.component
def test_assert_electroneutrality_get_property():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={'dynamic':False})
    m.fs.properties = DSPMDEParameterBlock(default={
        "solute_list": ["Ca_2+", "SO4_2-", "Na_+", "Cl_-", "Mg_2+"],
        "diffusivity_data": {("Liq", "Ca_2+"): 0.792e-9,
                             ("Liq", "SO4_2-"): 1.06e-9,
                             ("Liq", "Na_+"): 1.33e-9,
                             ("Liq", "Cl_-"): 2.03e-9,
                             ("Liq", "Mg_2+"): 0.706e-9},
        "mw_data": {"H2O": 18e-3,
                    "Na_+": 23e-3,
                    "Ca_2+": 40e-3,
                    "Mg_2+": 24e-3,
                    "Cl_-": 35e-3,
                    "SO4_2-": 96e-3},
        "stokes_radius_data": {"Na_+": 0.184e-9,
                               "Ca_2+": 0.309e-9,
                               "Mg_2+": 0.347e-9,
                               "Cl_-": 0.121e-9,
                               "SO4_2-": 0.230e-9},
        "charge": {"Na_+": 1,
                   "Ca_2+": 2,
                   "Mg_2+": 2,
                   "Cl_-": -1,
                   "SO4_2-": -2},
        "density_calculation": DensityCalculation.seawater,
        "activity_coefficient_model": ActivityCoefficientModel.davies
    })

    m.fs.stream = stream = m.fs.properties.build_state_block([0], default={'defined_state': True})

    mass_flow_in = 1 * pyunits.kg / pyunits.s
    feed_mass_frac = {'Na_+': 11122e-6,
                      'Ca_2+': 382e-6,
                      'Mg_2+': 1394e-6,
                      'SO4_2-': 2136e-6,
                      'Cl_-': 20300e-6}
    for ion, x in feed_mass_frac.items():
        mol_comp_flow = x * pyunits.kg / pyunits.kg * mass_flow_in / stream[0].mw_comp[ion]

        stream[0].flow_mol_phase_comp['Liq', ion].fix(mol_comp_flow)

    H2O_mass_frac = 1 - sum(x for x in feed_mass_frac.values())
    H2O_mol_comp_flow = H2O_mass_frac * pyunits.kg / pyunits.kg * mass_flow_in / stream[0].mw_comp['H2O']

    stream[0].flow_mol_phase_comp['Liq', 'H2O'].fix(H2O_mol_comp_flow)
    stream[0].temperature.fix(298.15)
    stream[0].pressure.fix(101325)

    assert not stream[0].is_property_constructed('pressure_osm')
    assert not stream[0].is_property_constructed('mass_frac_phase_comp')

    stream[0].assert_electroneutrality(defined_state=True,
                                       adjust_by_ion='Cl_-',
                                       get_property=['mass_frac_phase_comp', 'pressure_osm'])
    assert stream[0].is_property_constructed('mass_frac_phase_comp')
    assert stream[0].is_property_constructed('pressure_osm')
    assert not hasattr(stream, 'charge_balance')

    assert not stream[0].is_property_constructed('flow_vol')
    stream[0].assert_electroneutrality(defined_state=True,
                                       adjust_by_ion='Cl_-',
                                       get_property='flow_vol')
    assert stream[0].is_property_constructed('flow_vol')
    assert not hasattr(stream, 'charge_balance')

    # check that charge_balance constraint is deleted after failed solve
    stream[0].flow_vol_phase.fix(1e5)
    with pytest.raises(ValueError, match="The stateblock failed to solve while computing "
                                         "concentrations to check the charge balance."):
        stream[0].assert_electroneutrality(defined_state=True,
                                           adjust_by_ion='Cl_-')
    assert not hasattr(stream, 'charge_balance')

@pytest.mark.requires_idaes_solver
@pytest.mark.component
def test_assert_electroneutrality_get_property():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(default={'dynamic': False})
    m.fs.properties = DSPMDEParameterBlock(default={
        "solute_list": ["Ca_2+", "SO4_2-", "Na_+", "Cl_-", "Mg_2+"],
        "mw_data": {"H2O": 18e-3,
                    "Na_+": 23e-3,
                    "Ca_2+": 40e-3,
                    "Mg_2+": 24e-3,
                    "Cl_-": 35e-3,
                    "SO4_2-": 96e-3},
        "charge": {"Na_+": 1,
                   "Ca_2+": 2,
                   "Mg_2+": 2,
                   "Cl_-": -1,
                   "SO4_2-": -2},
        "density_calculation": DensityCalculation.seawater,
        "activity_coefficient_model": ActivityCoefficientModel.davies
    })
    m.fs.stream = stream = m.fs.properties.build_state_block([0], default={'defined_state': True})

    mass_flow_in = 1 * pyunits.kg / pyunits.s
    feed_mass_frac = {'Na_+': 11122e-6,
                      'Ca_2+': 382e-6,
                      'Mg_2+': 1394e-6,
                      'SO4_2-': 2136e-6,
                      'Cl_-': 20300e-6}
    for ion, x in feed_mass_frac.items():
        mol_comp_flow = x * pyunits.kg / pyunits.kg * mass_flow_in / stream[0].mw_comp[ion]

        stream[0].flow_mol_phase_comp['Liq', ion].fix(mol_comp_flow)

    H2O_mass_frac = 1 - sum(x for x in feed_mass_frac.values())
    H2O_mol_comp_flow = H2O_mass_frac * pyunits.kg / pyunits.kg * mass_flow_in / stream[0].mw_comp['H2O']

    stream[0].flow_mol_phase_comp['Liq', 'H2O'].fix(H2O_mol_comp_flow)
    stream[0].temperature.fix(298.15)
    stream[0].pressure.fix(101325)

    # check get_property works with tuple
    stream[0].assert_electroneutrality(defined_state=True,
                                       adjust_by_ion='Cl_-',
                                       get_property=('mass_frac_phase_comp', 'pressure_osm')
                                       )
    #check error when adjust_by_ion is not in solute list
    with pytest.raises(ValueError, match="adjust_by_ion must be set to the name of an "
                                         "ion in the list of solutes."):
        stream[0].assert_electroneutrality(defined_state=True,
                                           adjust_by_ion='foo')

    #check error when get_property is not None and defined_state is NOT true
    with pytest.raises(ValueError, match="Set defined_state to true if get_property"
                                         " = flow_mass_phase_comp"):
        stream[0].assert_electroneutrality(defined_state=False,
                                           get_property='flow_mass_phase_comp')

    #check error when state vars are fixed but defined_state is set to False
    with pytest.raises(AssertionError, match=re.escape("fs.stream[0].flow_mol_phase_comp[Liq,Ca_2+] was fixed. "
                                             "Either set defined_state=True or unfix flow_mol_phase_comp "
                                             "for each solute to check that electroneutrality is satisfied.")):
        stream[0].assert_electroneutrality(defined_state=False)

    # check error when get_property receives anything other than str, list, or tuple of strings
    with pytest.raises(TypeError,
                       match=re.escape("get_property must be a string or list/tuple of strings.")):
        stream[0].assert_electroneutrality(defined_state=True,
                                           adjust_by_ion='Cl_-',
                                           get_property=1)

    # check error when electroneutralit condition violated for stringent tolerance
    stream[0].flow_mol_phase_comp.unfix()
    with pytest.raises(AssertionError,
                       match=re.escape("Electroneutrality condition violated in fs.stream[0]. "
                                       "Ion concentrations should be adjusted to bring "
                                       "the result of 9.09E-13 closer towards 0.")):
        stream[0].assert_electroneutrality(defined_state=False,
                                           adjust_by_ion='Cl_-',
                                           tol=1e-18)
